/*
** $Id: conf.c,v 1.5 2003/01/10 00:58:41 pief Exp $
**
** MetalServe common stuff
** Copyright (c) 2002-2003 by Pieter Hollants <contact@metalserve.org>
**
** This program is licensed under the GNU General Public License, version 2,
** as published by the Free Software Foundation and available in the file
** COPYING and the Internet location http://www.gnu.org/licenses/gpl.html.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
** or FITNESS FOR A PARTICULAR PURPOSE.
**
** Configuration parser functions
**
** $Log: conf.c,v $
** Revision 1.5  2003/01/10 00:58:41  pief
** Introduced section support in the config file parser, which includes
** heavy changes to the interface and the operation of argument parser
** functions.
** Replaced malloc() calls with calloc() -- there's no reason not to use it.
** Added an argument parser functions for ranges of integer values.
** Extended copyright note to the new year.
**
** Revision 1.4  2002/12/01 23:33:16  pief
** This should have been revision 1.5.
**
** Revision 1.3  2002/11/24 11:01:43  pief
** Extended argparser_strlist() to support string values (strings that are
** mapped onto values instead of being stored themselves, eg. for "loglevel"
** configuration directive). In this context:
** - Added new field "params" to CFGDIRECTIVE structure for more complex
**   argument parser functions.
** - Added new CFGSTRVALUE structure for extended argparser_strlist()
**   functionality.
** - Renamed CFG_ERR_TOOMANYARGS to CFG_ERR_STRLISTARGCOUNT and added
**   CFG_ERR_EXCLARGVIOLATION.
** Renamed APFUNC_FLAG_xyz to DRTVFLAGS_xyz.
** Renamed CFG_FLAGS_xyz to CFGOPENFLAGS_xyz (and shortened IGNORE_UNKNOWN to
** IGNUNKNOWN).
**
** Revision 1.2  2002/09/24 22:05:13  pief
** For some reason I checked in ancient versions, so here are the current
** ones
**
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "base.h"
#include "strings.h"
#include "conf.h"
#include "conf-internal.h"

/*
** cfgerr(cfgfile)
**
** Sets the error code associated with the passed CFGFILE object and
** generates a suitable error string. 
*/
void cfgerr(CFGFILE *cfgfile, unsigned int cfgerrno)
{
	char errstrbuf[500];

	cfgfile->cfgerrno = cfgerrno;

	switch (cfgfile->cfgerrno)
	{
		case CFG_ERR_OUTOFMEM:
		{
			cfgfile->cfgerrstr = "Out of memory";
			return;
		}
		case CFG_ERR_INVDIRECTIVE:
		{
			sprintf(errstrbuf, "Unknown or invalid directive \"%s\"", cfgfile->dstr);
			cfgfile->cfgerrstr = errstrbuf;
			return;
		}
		case CFG_ERR_INVSECTION:
		{
			CFGSECTSTACKNODE *sectstacknode = (CFGSECTSTACKNODE *)cfgfile->sectstack->last;

			if (sectstacknode)
			{
				sprintf(errstrbuf, "Directive \"%s\" is not valid inside \"%s\" section (opened in line %d)",
				        cfgfile->d->name, sectstacknode->drtvname, sectstacknode->drtvlineno);
			}
			else
				sprintf(errstrbuf, "Directive \"%s\" is not valid in global section", cfgfile->d->name);

			cfgfile->cfgerrstr = errstrbuf;
			return;
		}
		case CFG_ERR_NOARG:
		{
			sprintf(errstrbuf, "No argument to directive \"%s\" specified", cfgfile->d->name);
			cfgfile->cfgerrstr = errstrbuf;
			return;
		}
		case CFG_ERR_ARGCOUNT:
		{
			sprintf(errstrbuf, "Invalid number of arguments to directive \"%s\" (min: %d, max: %d)",
				cfgfile->d->name, cfgfile->d->min, cfgfile->d->max);
			cfgfile->cfgerrstr = errstrbuf;
			return;
		}
		case CFG_ERR_ARGTOOSHORT:
		{
			sprintf(errstrbuf, "Argument to directive \"%s\" is too short (%d < %d characters)",
			        cfgfile->d->name, strlen(cfgfile->argstr), cfgfile->d->min);
			cfgfile->cfgerrstr = errstrbuf;
			return;
		}
		case CFG_ERR_ARGTOOLONG:
		{
			sprintf(errstrbuf, "Argument to directive \"%s\" is too long (%d > %d characters)",
			        cfgfile->d->name, strlen(cfgfile->argstr), cfgfile->d->max);
			cfgfile->cfgerrstr = errstrbuf;
			return;
		}
		case CFG_ERR_MISSMATCHQUOTES:
		{
			sprintf(errstrbuf, "Missing matching double quotes in argument to directive \"%s\"",
			        cfgfile->d->name);
			cfgfile->cfgerrstr = errstrbuf;
			return;
		}
		case CFG_ERR_INVARG:
		{
			sprintf(errstrbuf, "\"%s\" is no valid argument to directive \"%s\"",
			        cfgfile->argstr, cfgfile->d->name);
			cfgfile->cfgerrstr = errstrbuf;
			return;
		}
		case CFG_ERR_ARGTOOSMALL:
		{
			sprintf(errstrbuf, "Numeric argument to directive \"%s\" is too small (%d < %d)",
			        cfgfile->d->name, *((int *)cfgfile->d->target), cfgfile->d->min);
			cfgfile->cfgerrstr = errstrbuf;
			return;			
		}
		case CFG_ERR_ARGTOOLARGE:
		{
			sprintf(errstrbuf, "Numeric argument to directive \"%s\" is too large (%d > %d)",
			        cfgfile->d->name, *((int *)cfgfile->d->target), cfgfile->d->max);
			cfgfile->cfgerrstr = errstrbuf;
			return;			
		}
		case CFG_ERR_STRLISTARGCOUNT:
		{
			sprintf(errstrbuf, "Configuration parser supports at most %d arguments in string list to directive \"%s\"",
			        MAX_VALIST_PTRS, cfgfile->d->name);
			cfgfile->cfgerrstr = errstrbuf;
			return;			
		}
		case CFG_ERR_EXCLARGVIOLATION:
		{
			sprintf(errstrbuf, "\"%s\" violates previous exclusive argument to directive \"%s\"",
			        cfgfile->argstr, cfgfile->d->name);
			cfgfile->cfgerrstr = errstrbuf;
			return;
		}
		case CFG_ERR_INVDEFARG:
		{
			sprintf(errstrbuf, "Directive \"%s\" does not support the specifying of default properties",
			        cfgfile->d->name);
			cfgfile->cfgerrstr = errstrbuf;
			return;
		}
		case CFG_ERR_DUPLICATEARG:
		{
			sprintf(errstrbuf, "Duplicate occurance of argument \"%s\" to directive \"%s\"",
			        cfgfile->argstr, cfgfile->d->name);
			cfgfile->cfgerrstr = errstrbuf;
			return;
		}
		case CFG_ERR_MISSOPENINGBRACE:
		{
			sprintf(errstrbuf, "Missing opening brace for subsection of directive \"%s\"",
			        cfgfile->d->name);
			cfgfile->cfgerrstr = errstrbuf;
			return;
		}
		case CFG_ERR_OPENINGBRACEPOS:
		{
			cfgfile->cfgerrstr = "No opening brace allowed at this position";
			return;
		}
		case CFG_ERR_MISSCLOSINGBRACE:
		{
			CFGSECTSTACKNODE *opening_section = (CFGSECTSTACKNODE *)cfgfile->sectstack->last;

			sprintf(errstrbuf, "Missing closing brace for subsection of directive \"%s\" (opened in line %d)",
			        opening_section->drtvname, opening_section->drtvlineno);
			cfgfile->cfgerrstr = errstrbuf;
			return;
		}
		case CFG_ERR_CLOSINGBRACEPOS:
		{
			cfgfile->cfgerrstr = "No closing brace allowed at this position";
			return;
		}
		case CFG_ERR_NONE:
		default:
		{
			cfgfile->cfgerrstr = NULL;
		}
	}
}

/*
** cfgfile = cfgopen(filename, flags)
**
** Opens the specified configuration file for reading and returns a CFGFILE
** pointer for use with the other API functions.
**
** Returns TRUE on success or FALSE on failure, in which case the "errno"
** variable will contain the appropriate error code.
*/
CFGFILE *cfgopen(const char *filename, const int flags)
{
	CFGFILE *cfgfile;

	/* Allocate memory for CFGFILE structure */
	cfgfile = calloc(sizeof(CFGFILE), 1);
	if (!cfgfile)
	{
		errno = ENOMEM;
		return NULL;
	}

	/* Allocate memory for list of CFGSECTSTACKNODEs */
	cfgfile->sectstack = calloc(sizeof(LIST), 1);
	if (!cfgfile->sectstack)
	{
		free(cfgfile);
		errno = ENOMEM;
		return NULL;
	}

	/* Open the specified configuration file */
	cfgfile->fp = fopen(filename, "r");
	if (!cfgfile->fp)
	{
		free(cfgfile);
		return NULL;
	}

	/* Set flags and initialize CFGFILE object */
	cfgfile->flags           = flags;
	cfgfile->cfgerrno        = CFG_ERR_NONE;
	cfgfile->current_section = GLOBAL_SECTION;
	cfgfile->current_structp = NULL;

	return cfgfile;
}

/*
** cfgclose(cfgfile)
**
** Closes the specified configuration file and releases any resources used.
**
** After use, the variable cfgfile should be set to NULL, if chances are that
** it could be reused at a later time.
*/
void cfgclose(CFGFILE *cfgfile)
{
	if (!cfgfile) return;
	
	if (cfgfile->fp) fclose(cfgfile->fp);

	free(cfgfile->sectstack);
	free(cfgfile);
}

/*
** rc = cfgread(cfgfile, directives)
**
** Processes a configuration file using the supplied array of CFGDIRECTIVE
** structures.
**
** Returns TRUE on success or FALSE on failure, in which case _cfgerrno
** contains an error code for which a message may be obtained using
** cfgerror().
*/
BOOL cfgread(CFGFILE *cfgfile, CFGDIRECTIVE *directives)
{
	SECTOPEN_STATE_T open_subsection = MUST_NOT_OPEN;
	CFGSECTSTACKNODE *sectstacknode;

	/* Process line for line */
	while (fgets(cfgfile->linebuf, sizeof(cfgfile->linebuf), cfgfile->fp))
	{
		char *p, *q;

		cfgfile->lineno++;

		/* Strip whitespace, CRs and LFs */
		cfgfile->linep = trim_line(cfgfile->linebuf);

		/* Skip empty lines and comment lines */
		if ( (*cfgfile->linep == '#') || (*cfgfile->linep == '\0') )
			continue;

		/* Check whether line starts with an opening curled brace ('{') */
		if (cfgfile->linep[0] ==  '{')
		{
			/* Are there other characters in the line? */
			if (cfgfile->linep[1] != '\0')
			{
				cfgerr(cfgfile, CFG_ERR_INVARG);
				return FALSE;
			}

			/* Did the previously read directive allow for a section opening? */
			if (open_subsection == MUST_NOT_OPEN)
			{
				cfgerr(cfgfile, CFG_ERR_OPENINGBRACEPOS);
				return FALSE;
			}

			/* If so, insert previously allocated node into section stack */
			append_node(cfgfile->sectstack, (NODE *)sectstacknode);

			/* Just to be sure... */
			sectstacknode = NULL;

			open_subsection = MUST_NOT_OPEN;
			continue;
		}

		/* Check whether it should have opened one */
		if (open_subsection == MUST_OPEN)
		{
			cfgerr(cfgfile, CFG_ERR_MISSOPENINGBRACE);
			return FALSE;
		}

		/* Check whether it could have but didn't */
		if (open_subsection == CAN_OPEN)
		{
			/* Undo previously performed section change */
			cfgfile->current_section = sectstacknode->parent_section;
			cfgfile->current_structp = sectstacknode->parent_structp;

			/* Free allocated node */
			free(sectstacknode);
			sectstacknode = NULL;

			open_subsection = MUST_NOT_OPEN;
		}

		/* Check whether line starts with a closing curled brace ('}') */
		if (cfgfile->linep[0] == '}')
		{
			/* Are there other characters in the line? */
			if (cfgfile->linep[1] != '\0')
			{
				cfgerr(cfgfile, CFG_ERR_INVARG);
				return FALSE;
			}

			/* Is there a section to close or are we in the global section? */
			if (cfgfile->current_section == GLOBAL_SECTION)
			{
				cfgerr(cfgfile, CFG_ERR_CLOSINGBRACEPOS);
				return FALSE;
			}

			/* Pick up last node in section stack */
			sectstacknode = (CFGSECTSTACKNODE *)cfgfile->sectstack->last;	

			/* Restore section ID and structure pointer */
			cfgfile->current_section = sectstacknode->parent_section;
			cfgfile->current_structp = sectstacknode->parent_structp;

			/* Remove node from stack and free allocated memory */
			remove_node(cfgfile->sectstack, (NODE *)sectstacknode);
			free(sectstacknode);
			sectstacknode = NULL;

			continue;
		}

		/* Isolate directive */
		cfgfile->dstr = cfgfile->linep;
		cfgfile->argstr = cfgfile->linep + strcspn(cfgfile->linep, " \t\0");

		/* Skip whitespace to argument */
		if (*cfgfile->argstr != '\0')
		{
			*cfgfile->argstr = '\0';
			cfgfile->argstr++;
			cfgfile->argstr += strspn(cfgfile->argstr, " \t");
		}
		else cfgfile->argstr = NULL;

#if 0
		if (cfgfile->sectstack->count)
		{
			printf("%*cline %d section %d directive: [%s] arg: [%s]\n", cfgfile->sectstack->count,
			       ' ', cfgfile->lineno, cfgfile->current_section, cfgfile->dstr, cfgfile->argstr);
		}
		else
		{
			printf("line %d section %d directive: [%s] arg: [%s]\n", cfgfile->lineno,
			       cfgfile->current_section, cfgfile->dstr, cfgfile->argstr);
		}
#endif

		/* Look up directive */
		for (cfgfile->d = directives; cfgfile->d->name; cfgfile->d++)
		{
			if (!strcasecmp(cfgfile->dstr, cfgfile->d->name))
				break;
		}

		/* Did we get a match? */
		if (cfgfile->d->name)
		{

			/* Determine whether this directive may (or must) open a subsection */
			if (cfgfile->d->target)
				open_subsection = MUST_NOT_OPEN;
			else if (cfgfile->d->min)
				open_subsection = MUST_OPEN;
			else
				open_subsection = CAN_OPEN;
		}
		else
		{
			/* Check whether cfgopen() was called with CFGOPENFLAGS_IGNUNKNOWN set */
			if (cfgfile->flags & CFGOPENFLAGS_IGNUNKNOWN)
			{
				/* If so, we treat unknown directives as if they can open subsections */
				open_subsection = CAN_OPEN;
			}
			else
			{
				/* If not, raise the appropriate error */
				cfgerr(cfgfile, CFG_ERR_INVDIRECTIVE);
				return FALSE;
			}
		}

		/* Check whether no argument was specified */
		if (!cfgfile->argstr)
		{
			cfgerr(cfgfile, CFG_ERR_NOARG);
			return FALSE;	
		}

		/* Closing curled brace ('}') detection. We can't just strchr() here,
		   because the brace might be escaped (prefixed with a backslash '\') */
		p = cfgfile->argstr;
		do
		{
			/* Locate closing curled braces within remaining part of argument */
			if (q = strchr(p, '}'))
			{
				/* Check whether this brace was escaped */
				if ( (q != cfgfile->argstr) && (*(q-1) == '\\') )
				{
					/* Continue search in remainder of argument */
					p = q;
					continue;
				}

				/* Closing curled braces are not allowed as part of the
				   argument string and must always appear on a seperate line */
				cfgerr(cfgfile, CFG_ERR_CLOSINGBRACEPOS);
				return FALSE;
			}
		}
		while (q);

		/* Opening curled brace ('{') detection. Again, we can't just strchr()
		   here for the same reasons as above, however we also accept an opening
		   brace as the last character in the argument here. */
		p = cfgfile->argstr;
		do
		{
			/* Locate opening curled braces within remaining part of argument */
			if (q = strchr(p, '{'))
			{
				/* Check whether this brace was escaped */
				if ( (q != cfgfile->argstr) && (*(q-1) == '\\') )
				{
					/* Continue search in remainder of argument */
					p = q;
					continue;
				}

				/* Check whether this brace is the last character and
				   whether the just-read directive allows for a section
				   opening */
				if ( (*(q+1) != '\0') || (open_subsection != MUST_NOT_OPEN) )
				{
					cfgerr(cfgfile, CFG_ERR_OPENINGBRACEPOS);
					return FALSE;
				}

				/* We found an unescaped opening brace as the last character
				   in the argument, so we can complete the section change
				   almost immediately. */
				open_subsection = DOES_OPEN;

				/* Remove the opening brace from the argument, so that the
				   argparser for the current directive can operate correctly. */
				*q = '\0';

				/* We don't need to search no more */
				break;
			}
		}
		while (q);

		/* Save section ID and structure pointer in a CFGSECTSTACKNODE structure,
		   unless we absolutely know that we'll stay in the current section */
		if (open_subsection != MUST_NOT_OPEN)
		{
			/* Allocate memory for CFGSECTSTACKNODE */
			sectstacknode = calloc(sizeof(CFGSECTSTACKNODE), 1);
			if (!sectstacknode)
			{
				cfgerr(cfgfile, CFG_ERR_OUTOFMEM);
				return FALSE;
			}

			/* Backup section ID and structure pointer */
			sectstacknode->parent_section = cfgfile->current_section;
			sectstacknode->parent_structp = cfgfile->current_structp;

			/* Remember directive that opened this subsection */
			if (cfgfile->d->name)
				sectstacknode->drtvname = cfgfile->d->name;
			else
				sectstacknode->drtvname = cfgfile->dstr;
			sectstacknode->drtvlineno = cfgfile->lineno;
		}

		/* Only if the directive is known, we check its validity and call its argument
		   parsing function */
		if (cfgfile->d->name)
		{
			void *target;

			/* Check whether directive is valid in current section */
			if (!(cfgfile->d->valid_sections & cfgfile->current_section))
			{
				cfgerr(cfgfile, CFG_ERR_INVSECTION);
				return FALSE;
			}

			/* If this is a SUBSESSION_DIRECTIVE, let "target" point to the structp
			   (structure base pointer) to be modified. Argument parsing functions may,
			   in their calculation of a new structure base pointer, read the value
			   of *target as it will contain the current (or then, parent) section's
			   structure base pointer. */
			if (!cfgfile->d->target)
				target = &cfgfile->current_structp;

			/* Otherwise calculate "target" pointer */
			else
			{
				/* Global directives modify global variables directly specified in
				   the "target" field */
				if (cfgfile->current_section == GLOBAL_SECTION)
					target = cfgfile->d->target;

				/* Session-dependant directives modify parts of the structure referenced
				   by cfgfile->current_structp */
				else
				{
					target = cfgfile->current_structp + (size_t)cfgfile->d->target;
				}
			}

#if 0
			printf("current_structp: %p  offset: %d  target: %p  *target: %p\n",
				cfgfile->current_structp, (size_t)cfgfile->d->target, target, *((int *)target));
#endif

			/* Call the argument parsing function */
			cfgfile->cfgerrno = cfgfile->d->apfunc(cfgfile->d, cfgfile->current_section, target, &cfgfile->argstr);
			if (cfgfile->cfgerrno != CFG_ERR_NONE)
			{
				cfgerr(cfgfile, cfgfile->cfgerrno);
				return FALSE;
			}
		}

		/* Unless we absolutely know that we'll stay in the current section,
		   pro-actively perform section change. */
		if (open_subsection != MUST_NOT_OPEN)
		{
			/* Change section ID */
			if (cfgfile->d->name)
				cfgfile->current_section = cfgfile->d->max;
			else
				cfgfile->current_section = UNKNOWN_SECTION;

			/* It is assumed that the argument parser function will
			   let cfgfile->structp point to some base structure */
		}
	}
	if (errno) return FALSE;

	/* Check whether all subsections were closed properly */
	if (cfgfile->sectstack->last)
	{
		cfgerr(cfgfile, CFG_ERR_MISSCLOSINGBRACE);
		return FALSE;
	}

	return TRUE;
}

/*
** cfgerrno = cfgerrno(CFGFILE *cfgfile)
**
** Returns the error code of the last configuration error that occured while
** reading the specified configuration file, which may be CFG_ERR_NONE.
*/
int cfgerrno(CFGFILE *cfgfile)
{
	return cfgfile->cfgerrno;
}

/*
** cfgerrstr = cfgerrstr(CFGFILE *cfgfile)
**
** Returns a string describing the last configuration error that occured while
** reading the specified configuration file, which may be a NULL pointer if
** no error occured.
*/
char *cfgerrstr(CFGFILE *cfgfile)
{
	return cfgfile->cfgerrstr;
}

/*
** lineno = cfglineno(CFGFILE *cfgfile)
**
** Returns the number of the last line that was read before a configuration
** error occured.
*/
int cfglineno(CFGFILE *cfgfile)
{
	return cfgfile->lineno;
}

/*
** Argument parser function for C-style strings (terminated with a zero byte).
**
** "min" and "max" specify the minimum and maximum allowed string length.
*/
APFUNC(argparser_string)
{
	char *arg = *argp;
	char *unescquoted_string;

	/* Check for empty string */
	if ( (*arg == '"') && (*(arg+1) == '"') )
	{
		/* Are empty strings allowed? */
		if (current->min == 0)
		{
			*((char **)target) = NULL;
			return CFG_ERR_NONE;
		}
		else
			return CFG_ERR_NOARG;
	}

	/* Check argument length */
	if (strlen(arg) < current->min)
		return CFG_ERR_ARGTOOSHORT;
	if (strlen(arg) > current->max)
		return CFG_ERR_ARGTOOLONG;

	/* Assign copy of "unescquoted" string argument */
	unescquoted_string = unescquote(arg);
	if (unescquoted_string < 0)
		return CFG_ERR_OUTOFMEM;
	else if (!unescquoted_string)
		return CFG_ERR_MISSMATCHQUOTES;

	*((char **)target) = unescquoted_string;

	return CFG_ERR_NONE;
}

/*
** Argument parser function for switches. The keywords "yes", "on" and "true"
** cause the configuration variable to be set to 1, the keywords "no", "off"
** and "false" set it to 0. Other arguments cause the parser to fail.
*/
APFUNC(argparser_switch)
{
	char *arg = *argp;

	/* Check for recognized keywords */
	if ( (strcmp(arg, "yes") == 0) || (strcmp(arg, "on") == 0) || (strcmp(arg, "true") == 0) )
	{
		*((BOOL *)target) = TRUE;
		return CFG_ERR_NONE;
	}
	else if ( (strcmp(arg, "no") == 0) || (strcmp(arg, "off") == 0) || (strcmp(arg, "false") == 0) )
	{
		*((BOOL *)target) = FALSE;
		return CFG_ERR_NONE;
	}

	/* An unrecognized keyword was specified */
	return CFG_ERR_INVARG;
}

/*
** Argument parser for string lists (lists with a variable number of
** space-seperated strings)
**
** "min" and "max" specify the minimum required and maximum allowed number of
** arguments.
**
** If "params" is NULL, the "target" variable will point to an array of
** pointers to each string, which must be free()'d by the calling application.
**
** If "params" is non-NULL, it must point to an array of CFGSTRVALUE structures,
** terminated with the LAST_CFGSTRVALUE macro, that will be used to map the
** extracted strings onto specific values, which will be ORed ("exclusive" ==
** FALSE) or overwrite each other ("exclusive" == TRUE). An unrecognized string
** value will be honoured with a CFG_ERR_INVARG.
*/
APFUNC(argparser_stringlist)
{
	char *arg = *argp, *nextarg, *unescquoted_arg;
	BOOL quoted, escaped;
	int strings = 0;

        char **ptr_array, **ptr;
	int ptrs_left;

	BOOL excl_strvalue_found = FALSE;
	int val = 0;

	/* Unless we deal with strvalues, we need to store the extracted strings somewhere */
	if (!current->params)
	{
		/* Allocate pointer array for extracted strings */
		ptr_array = calloc(MAX_VALIST_PTRS + 1, sizeof(char *));
		if (!ptr_array)
			return CFG_ERR_OUTOFMEM;
		ptrs_left = MAX_VALIST_PTRS;
		ptr = ptr_array;
		*((char ***)target) = ptr_array;
	}

	/* Process arguments */
	for (nextarg = arg; arg; arg = nextarg)
	{
		/* Determine next string in list or abort if the end of the line is reached */
		quoted = FALSE;
		escaped = FALSE;
		while (*nextarg != '\0')
		{
			/* At THIS place, if we are in escape mode, we don't really have to care
			   about the second part of an escape sequence, we just reset "escaped" */
			if (escaped)
				escaped = FALSE;
			else
			{
				/* "Space" and "Tab" are recognized as argument seperators.
				   While "Space" is not recognized when in quoted mode, Tab
				   always is. */
				if ( ( (!quoted) && (*nextarg == ' ') ) || (*nextarg == '\t') )
					break;

				/* Toggle quote mode? */
				if (*nextarg == '"')
				{
					if (quoted)
						quoted = FALSE;
					else
						quoted = TRUE;
				}

				/* If this is the start of an escape sequence, make sure that
				   the next character is not interpreted. Note that we don't
				   just increase nextarg because we might skip the final null
				   byte (this way, the while-condition will trigger). */
				if (*nextarg == '\\')
					escaped = TRUE;
			}

			nextarg++;
		}

		if (*nextarg != '\0')
		{
			*nextarg = 0;
			nextarg++;
		}
		else nextarg = NULL;

		/* Create "unescquoted" copy of this argument */
		unescquoted_arg = unescquote(arg);
		if (unescquoted_arg < 0)
			return CFG_ERR_OUTOFMEM;
		else if (!unescquoted_arg)
			return CFG_ERR_MISSMATCHQUOTES;

		strings++;

		/* If we deal with strvalues, try to find a matching value */
		if (current->params)
		{
			CFGSTRVALUE *strvalue;
			BOOL found = FALSE;

			/* Match case-insensitively against known strings in CFGSTRVALUE structure array */
			for (strvalue = (CFGSTRVALUE *)current->params; (!found) && (strvalue->string); strvalue++)
			{
				if (!strcasecmp(unescquoted_arg, strvalue->string))
				{
					found = TRUE;
					break;
				}
			}

			/* Error handling if the string wasn't found or follows an exclusive STRVALUE */
			if ( (!found) || (excl_strvalue_found) )
			{
				/* We must update argp (which is cfgfile->argstr) since increasing "arg"
				   in the for loop above just increased the value of a local copy of
				   argstr, and the error message would always be generated for the first
				   string. */
				*argp = unescquoted_arg;

				if (!found)
					return CFG_ERR_INVARG;
				else
					return CFG_ERR_EXCLARGVIOLATION;
			}

			/* If the matching CFGSTRVALUE has the exclusive flag set, blindly set the
			   target variable to the associated value, overwriting previous values.
			   Otherwise OR it with the existing value. */
			if (strvalue->exclusive)
			{
				val = strvalue->value;
				excl_strvalue_found = TRUE;
			}
			else
				val |= strvalue->value;
		}
		/* Otherwise store in pointer array */
		else
		{
			*ptr = unescquoted_arg;
			ptr++;
			ptrs_left--;

			/* If all pointers have been used, bail out and error */
			if (!ptrs_left)
				return CFG_ERR_STRLISTARGCOUNT;
		}
	}

	/* If we deal with strvalues, store the calculated value. Otherwise store an
	   additional NULL pointer so that calling functions know where the pointer array ends */
	if (current->params)
		*((int *)target) = val;
	else
		*ptr = NULL;

	/* Check whether enough/too many arguments were specified */
	if (strings < current->min || strings > current->max)
		return CFG_ERR_ARGCOUNT;

	return CFG_ERR_NONE;
}

/*
** Argument parser function for integer values.
**
** "min" and "max" specify minimum and maximum values.
*/
APFUNC(argparser_int)
{
	char *arg = *argp, *p;
	int value;

	/* Attempt to convert string representation to a numeric value */
	value = strtol(arg, &p, 10);
	if (*p != '\0')
		return CFG_ERR_INVARG;

	/* Store converted value already here so that the following checks
	   have the value available in their error messages */
	*((int *)target) = value;

	/* Check whether converted value is within the allowed range */
	if (value < current->min)
		return CFG_ERR_ARGTOOSMALL;
	if (value > current->max)
		return CFG_ERR_ARGTOOLARGE;

	return CFG_ERR_NONE;
}

/*
** Argument parser function for integer ranges.
**
** This function expects an argument that consists of two integer values,
** seperated by a hyphen ('-'). The first value is considered as the start
** of a range, the second value as its end. The first value must be lower
** than or equal to the second value. If the argument consists of a single
** integer only (that is, no hyphen and no second value follows), the
** parser interprets it as 'value-value'.
**
** "target" must point to the integer value for the START of the range. The
** argparser will place the END of the range in "target"+1, where +1 means
** the immediately following "int" variable (read: "int", not "short", not
** "char"). Don't wonder that your variables get thrashed if you don't keep
** this in mind!
**
** "min" specifies the minimum value for the start and "max" the maximum
** value for the end of the range.
*/
APFUNC(argparser_intrange)
{
	char *arg = *argp, *p;
	int start, end, *endp;

	/* Attempt to convert first part of string to a numeric value */
	start = end = strtol(arg, &p, 10);

	/* Store converted value as start of range */
	*((int *)target) = start;

	/* Check whether converted value is within the allowed range */
	if (start < current->min)
		return CFG_ERR_ARGTOOSMALL;

	/* Is this the only number? */
	if (*p == '-')
	{
		/* Skip hyphen ('-') */
		p++;

		/* Attempt to convert rest of string to a numeric value */
		end = strtol(p, &p, 10);
	}
	else if (*p != '\0')
		return CFG_ERR_INVARG;

	/* Store value as end of range, too */
	*(((int *)target)+1)=end;;

	/* Check whether converted value is within the allowed range */
	if (end > current->max)
		return CFG_ERR_ARGTOOLARGE;

	/* Check whether start of range is smaller than end of range */
	if (start > end)
		return CFG_ERR_INVARG;

	return CFG_ERR_NONE;
}
