/*
** $Id: conf.h,v 1.6 2003/01/10 00:58:41 pief Exp $
**
** MetalServe common stuff
** Copyright (c) 2002-2003 by Pieter Hollants <contact@metalserve.org>
**
** This program is licensed under the GNU General Public License, version 2,
** as published by the Free Software Foundation and available in the file
** COPYING and the Internet location http://www.gnu.org/licenses/gpl.html.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
** or FITNESS FOR A PARTICULAR PURPOSE.
**
** Header file for configuration parser functions
**
** $Log: conf.h,v $
** Revision 1.6  2003/01/10 00:58:41  pief
** Introduced section support in the config file parser, which includes
** heavy changes to the interface and the operation of argument parser
** functions.
** Replaced malloc() calls with calloc() -- there's no reason not to use it.
** Added an argument parser functions for ranges of integer values.
** Extended copyright note to the new year.
**
** Revision 1.5  2002/12/01 23:32:37  pief
** Removed superfluos argparser_command().
** Removed "flags" field from CFGDIRECTIVE structure: argparser_string now
** never unescapes and unquotes whereas the arguments in a string list are
** always unescquote()'d.
**
** Revision 1.4  2002/11/24 11:01:43  pief
** Extended argparser_strlist() to support string values (strings that are
** mapped onto values instead of being stored themselves, eg. for "loglevel"
** configuration directive). In this context:
** - Added new field "params" to CFGDIRECTIVE structure for more complex
**   argument parser functions.
** - Added new CFGSTRVALUE structure for extended argparser_strlist()
**   functionality.
** - Renamed CFG_ERR_TOOMANYARGS to CFG_ERR_STRLISTARGCOUNT and added
**   CFG_ERR_EXCLARGVIOLATION.
** Renamed APFUNC_FLAG_xyz to DRTVFLAGS_xyz.
** Renamed CFG_FLAGS_xyz to CFGOPENFLAGS_xyz (and shortened IGNORE_UNKNOWN to
** IGNUNKNOWN).
**
** Revision 1.3  2002/10/25 15:59:37  pief
** Replaced names of per-file #define's to avoid possible namespace clashes
**
** Revision 1.2  2002/09/24 22:05:13  pief
** For some reason I checked in ancient versions, so here are the current
** ones
**
*/

#ifndef _METALSERVE_CONF_H
#define _METALSERVE_CONF_H

/*
** The CFGFILE structure is for internal use only, but is declared here because
** we need it in the prototypes.
*/
typedef struct _cfgfile CFGFILE;

/*
** Forward declaration of CFGDIRECTIVE structure
*/
typedef struct _cfgdirective CFGDIRECTIVE;

/*
** Argument parser functions receive the following arguments, which are hidden by the
** macro APFUNC_ARGS:
**
** - "current": the currently processed directive (eg. to obtain argument restrictions)
** - "section": the ID of the currently active section (eg. GLOBAL_SECTION)
** - "target" : a pointer to the the target variable to be modified. If this is a SESSION_DIRECTIVE,
**              "target" must be set to the structure base pointer upon which the subsection's
**              directives will operate (reading "target" beforehand will yield the structure
**              base pointer of the parent section).
** - "argp"   : a pointer to a pointer to the remainder of the config line (ie. *arg is the
**              the argument to be parsed, arg itself may be advanced by the function)
*/
#define APFUNC_ARGS CFGDIRECTIVE *current, unsigned short section, const void *target, char **argp

/*
** They return one of the CFG_ERR_* error codes.
**
** These semantics are hidden behind the macro APFUNC which should be used for the
** declaration and definition of argument parser functions.
*/
#define APFUNC(_name) unsigned int _name(APFUNC_ARGS)

/*
** These are the prototypes for the built-in argument parser functions.
*/
APFUNC(argparser_string);
APFUNC(argparser_switch);
APFUNC(argparser_stringlist);
APFUNC(argparser_int);
APFUNC(argparser_intrange);

/*
** A CFGDIRECTIVE structure specifies a "known" configuration directive. The structure specifies:
** - a (generic) argument parser function, which takes care of interpreting the argument
** - the target variable to be modified (used by the apfunc)
** - min and max restrictions that restrict valid arguments (used by the apfunc)
** - a bitmask specifying restrictions as to in which sections this directive is valid, and
** - a pointer to an additional parameter structure (also used by apfunc)
*/
struct _cfgdirective
{
	char		*name;					/* Name of directive */
	unsigned int	(*apfunc)(APFUNC_ARGS);			/* Function responsible for argument parsing */
	void		*target;				/* Target variable to be modified by apfunc */
	int		min, max;				/* Min/Max string length/value/#arguments */
	unsigned short	valid_sections;				/* Bitmask of sections in which this directive is valid */
	void		*params;				/* Pointer to some parameter structure (set to NULL if unused) */
};

/*
** Default #define's for the "section" field above. Applications should define their own
** macros with appropriate bits (2^x with x>0) to restrict directives to certain (sub-)sections.
*/
#define GLOBAL_SECTION	1					/* Valid outside of subsections only */
#define UNKNOWN_SECTION ~GLOBAL_SECTION				/* For LAST_CFGDIRECTIVE and unknown directives */

/*
** Directives that open a subsection should be defined using the following macro. 
**
** The configuration parser recognizes that a directive can open a subsection when it
** doesn't modify any variable (ie. "target" is NULL). In this case, "min" is interpreted as a
** boolean value specifying whether the directive MUST open a subsection or just CAN and "max"
** specifies the ID for the subsection to be opened. "apfunc" is supposed to set cfgfile->structp
** to point to the base address of a structure to be modified. The config parser will interpret
** "target" to be an offset which is to be added to this base address in order to determine the
** actual variable to be modified.
*/
#define SUBSECTION_DIRECTIVE(name, apfunc, valid_sections, must_open, target_section) \
 { name, apfunc, NULL, must_open, target_section, valid_sections, NULL }

/*
** Directives that are only valid within certain subsections often want to modify a component
** of some structure instead of a global variable. For this purpose, we provide the offsetof()
** macro, which, although defined by ANSI C, is unfortunately not universally available. Note
** that while common definitions define this as (size_t), we use (void *) since that is the type
** of the structure field the offset will be stored in.
*/
#undef offsetof
#define offsetof(struct_type, member) ((void *)&(((struct_type *)NULL)->member))

/*
** An array of CFGDIRECTIVE structures defines the entire configuration file. It must be terminated
** with the LAST_CFGDIRECTIVE macro.
*/
#define LAST_CFGDIRECTIVE { NULL, NULL, NULL, 0, 0, UNKNOWN_SECTION, NULL }

/*
** Maximum number of entries in variable-argument lists such as string lists
*/
#define MAX_VALIST_PTRS 30

/*
** A CFGSTRVALUE structure contains symbolic strings that may appear as arguments to a configuration
** directive and are mapped by argparser_strvalue onto a specific value.
*/
typedef struct _cfgstrvalue
{
	char		*string;				/* String to be matched */
	int		value;					/* Value for target variable */
	BOOL		exclusive;				/* Is this an exclusive value? */
} CFGSTRVALUE;

/*
** An array of CFGSTRVALUE structure defines the allowed symbolic strings for a configuration
** directive. It must be terminated with the LAST_CFGSTRVALUE macro.
*/
#define LAST_CFGSTRVALUE { NULL, 0, FALSE }

/*
** Configuration error code definitions (specify these in your argument parser functions as argument
** to cfgerr() calls)
*/
enum _cfgerrcodes
{
	CFG_ERR_NONE,					/* No error */
	CFG_ERR_OUTOFMEM,				/* Out of memory */
	CFG_ERR_INVDIRECTIVE,				/* Invalid/unknown directive */
	CFG_ERR_INVSECTION,				/* Directive is not valid in current section */
	CFG_ERR_NOARG,					/* No argument specified */
	CFG_ERR_ARGCOUNT,				/* Invalid number of arguments */
	CFG_ERR_ARGTOOSHORT,				/* Argument too short */
	CFG_ERR_ARGTOOLONG,				/* Argument too long */
	CFG_ERR_MISSMATCHQUOTES,			/* Missing matching double quotes */
	CFG_ERR_INVARG,					/* Invalid argument */
	CFG_ERR_ARGTOOSMALL,				/* Numeric argument too small */
	CFG_ERR_ARGTOOLARGE,				/* Numeric argument too large */
	CFG_ERR_STRLISTARGCOUNT,			/* Too many arguments in a string list */
	CFG_ERR_EXCLARGVIOLATION,			/* Argument violates previous exclusive argument */
	CFG_ERR_INVDEFARG,				/* Directive doesn't support "default" argument */
	CFG_ERR_DUPLICATEARG,				/* Argument to directive was already specified */
	CFG_ERR_MISSOPENINGBRACE,			/* Missing open brace */
	CFG_ERR_OPENINGBRACEPOS,			/* Opening brace at illegal position */
	CFG_ERR_MISSCLOSINGBRACE,			/* Missing closing brace */
	CFG_ERR_CLOSINGBRACEPOS				/* Closing brace at illegal position */
};

/*
** Main API prototypes.
*/
CFGFILE *cfgopen(const char *filename, const int flags);
void cfgclose(CFGFILE *cfgfile);
BOOL cfgread(CFGFILE *cfgfile, CFGDIRECTIVE *directives);
int cfgerrno(CFGFILE *cfgfile);
char *cfgerrstr(CFGFILE *cfgfile);
int cfglineno(CFGFILE *cfgfile);

/*
** Flags to the cfgopen() function
*/
#define CFGOPENFLAGS_NONE	0			/* No flags set */
#define CFGOPENFLAGS_IGNUNKNOWN	1			/* Ignore unknown directives (instead of reporting errors) */

#endif
