/*
** $Id: list.c,v 1.3 2003/01/10 00:51:11 pief Exp $
**
** MetalServe common stuff
** Copyright (c) 2002-2003 by Pieter Hollants <contact@metalserve.org>
**
** This program is licensed under the GNU General Public License, version 2,
** as published by the Free Software Foundation and available in the file
** COPYING and the Internet location http://www.gnu.org/licenses/gpl.html.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
** or FITNESS FOR A PARTICULAR PURPOSE.
**
** List functions
**
** $Log: list.c,v $
** Revision 1.3  2003/01/10 00:51:11  pief
** Extended copyright note to the new year
** Replaced malloc() with calloc() -- there's no reason not to use the latter
**
** Revision 1.2  2002/09/24 22:05:13  pief
** For some reason I checked in ancient versions, so here are the current
** ones
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>

#include "list.h"

/*
** list = allocate_list()
**
** Allocates and initializes a new LIST list header structure.
**
** Returns a pointer to a LIST structure or NULL on failure.
*/
LIST *allocate_list(void)
{
	return calloc(sizeof(LIST), 1);
}

/*
** free_list(list, func)
**
** Traverses "list" in a way that allows for "func" to free the single
** nodes. After all nodes have been freed this way, the list itself is
** freed.
*/
void free_list(LIST *list, free_node_func func)
{
	NODE *node, *next_node;

	for (node=list->first; node; node=next_node)
	{
		next_node=node->next;
		func(node);
	}

	free(list);
}

/*
** traverse_list(list, func)
**
** Traverses "list" and calls "func" for each node in the list. Continues
** traversing as long as "func" returns zero.
**
** Returns the current node if "func" returned a non-zero value for
** this node or NULL if the end of "list" has been reached, or "list" is
** NULL.
*/
NODE *traverse_list(LIST *list, traverse_node_func func)
{
	NODE *node;
	
	for (node=list->first; node; node=node->next)
	{
		if (func(node)) return node;
	}
	
	return NULL;
}

/*
** insert_node(list, node, new_node)
**
** Inserts "new_node" behind "node" in "list". If "node" is NULL,
** "new_node" is inserted in front of the first node in the list.
*/
void insert_node(LIST *list, NODE *node, NODE *new_node)
{
	/* Empty list? */
	if (!list->last)
	{
		new_node->next = NULL;
		new_node->prev = NULL;
		list->first    = new_node;
		list->last     = new_node;
	}
	/* OK, there's at least one element. Insert at beginning of list? */
	else if (node == NULL)
	{
		new_node->next    = list->first;
		list->first->prev = new_node;
		new_node->prev    = NULL;
		list->first       = new_node;
	}
	/* Insert as 2nd ... last element */
	else
	{
		new_node->next = node->next;
		if (node->next) node->next->prev = new_node;
		new_node->prev = node;
		node->next     = new_node;
		if (list->last == node) list->last = new_node;
	}

	list->count++;
}

/*
** insert_prionode(list, new_prionode)
**
** Inserts the PRIONODE structure "new_prionode" into the priority-ordered
** list "list". "new_prionode" will be added behind the last PRIONODE of
** equal priority.
*/
void insert_prionode(LIST *list, PRIONODE *new_prionode)
{
	PRIONODE *current, *next = NULL;

	/* If the list is empty, we can just append to it */
	if (!list->first)
	{
		append_node(list, (NODE *)new_prionode);
		return;
	}

	/* Check top position */
	current = (PRIONODE *)list->first;
	if (new_prionode->priority > current->priority)
	{
		insert_node(list, NULL, (NODE *)new_prionode);
		return;
	}

	/* No match, so let's find a suitable place */
	while (current->next)
	{
		next = (PRIONODE *)current->next;

		if ( (next) && (new_prionode->priority > next->priority) )
		 break;
		     
		current = next;
	}

	insert_node(list, (NODE *)current, (NODE *)new_prionode);
}

/*
** rc = remove_node(list, node)
**
** Removes "node" from "list". Does not check if "node" is actually a
** member of "list" (why should we spend O(n) checking that...)
*/
void remove_node(LIST *list, NODE *node)
{
	if (node->prev)
	{
		if (node->next)
		{
			node->prev->next = node->next;
			node->next->prev = node->prev;
		}
		else
		{
			node->prev->next = NULL;
			list->last       = node->prev;
		}
	}
	else
	{
		if (node->next)
		{
			node->next->prev = NULL;
			list->first      = node->next;
		}
		else
		{
			list->first      = NULL;
			list->last       = NULL;
		}
	}
	
	list->count--;
}
