/*
** $Id: log.c,v 1.4 2003/01/10 00:50:08 pief Exp $
**
** MetalServe common stuff
** Copyright (c) 2002-2003 by Pieter Hollants <contact@metalserve.org>
**
** This program is licensed under the GNU General Public License, version 2,
** as published by the Free Software Foundation and available in the file
** COPYING and the Internet location http://www.gnu.org/licenses/gpl.html.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
** or FITNESS FOR A PARTICULAR PURPOSE.
**
** Logging functions
**
** $Log: log.c,v $
** Revision 1.4  2003/01/10 00:50:08  pief
** Extended copyright note to the new year
**
** Revision 1.3  2002/12/09 16:18:22  pief
** Moved output to stderr up in do_logmsg() since we otherwise output a zero
** byte-terminated string which is intended to be written to the logfile only
**
** Revision 1.2  2002/11/24 10:51:53  pief
** Fixed parameter name in function description of logmsg()
**
** Revision 1.1  2002/10/25 16:09:17  pief
** Added logging routines
**
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#define SYSLOG_NAMES
#include <syslog.h>
#ifndef INTERNAL_NOPRI
#include "../lib/syslog.h"
#endif
#undef SYSLOG_NAMES
#include <time.h>

#include "base.h"
#include "log.h"

static FILE *_logfd;
static int _loglevel;

extern char *_prgname;

/*
** loglevel = getloglevelval(name)
**
** Returns the integer value for the specified loglevel name or -1 if no
** valid loglevel was specified.
*/
int getloglevelval(const char *name)
{
	CODE *priority;
	
	for (priority = prioritynames; priority->c_name; priority++)
	{
		if (!strcmp(name, priority->c_name))
			return priority->c_val;	
	}

	return -1;
}

/*
** name = getloglevelname(loglevel)
**
** Returns the loglevel name associated with "loglevel" or NULL if the loglevel
** specified was invalid.
*/
const char *getloglevelname(const int loglevel)
{
	CODE *priority;
	
	for (priority = prioritynames; priority->c_name; priority++)
	{
		if (loglevel == priority->c_val)
			return priority->c_name;	
	}

	return NULL;
}

/*
** rc = logopen(logfile_name, loglevel, stdout)
**
** Opens the specified logfile "logfile_name" for writing. Only messages with
** a severity <= "loglevel" will be written.
**
** Returns TRUE on success and FALSE on failure.
*/
BOOL logopen(const char *logfile_name, const int loglevel)
{
	_logfd = fopen(logfile_name, "a");
	if (!_logfd)
		return FALSE;

	_loglevel = loglevel;

	return TRUE;
}

/*
** logclose()
**
** Closes the logfile.
*/
void logclose(void)
{
	if (_logfd)
	{
		fclose(_logfd);
		_logfd = NULL;
	}
}

/*
** do_logmsg(file, func, line, severity, stderr, fmt, ...)
**
** Writes a log message to the logfile using "fmt" as format and any additionally
** supplied arguments. "severity" is one of the standard syslog priority levels,
** however LOG_EMERG and LOG_ALERT should not be used as there is nothing inside.
** "severity" must be less than or equal to the loglevel specified with the logopen
** call. "errmsg" specifies whether the message is an error message.
**
** Messages are prefixed with the calling program's name and, if the loglevel
** is set to LOG_DEBUG, with the C source file name, the function name and the
** appropriate line within the C Source file. Messages written to the logfile
** get a timestamp as well, which is omitted from stderr for aesthetical reasons.
**
** Never call do_logmsg() directly -- use the macros logmsg() and logerr() from
** log.h instead.
*/
void do_logmsg(const char *file, const char *func, const int line, const int severity,
               const BOOL errmsg, const char *fmt, ...)
{
	time_t now;
	static char timestampbuf[LOG_TIMESTAMPBUF_LENGTH];

	static char outputbuf[LOG_OUTPUTBUF_LENGTH];
	char *p, *linep;
	va_list va;

	/* Check if a logfile was opened and compare severity with configured loglevel */
	if ( (!_logfd) || (severity > _loglevel) ) return;

	/* Generate timestamp for messages written to the logfile */
	now = time(NULL);
	strftime(timestampbuf, LOG_TIMESTAMPBUF_LENGTH, LOG_TIMESTAMP_FORMAT, localtime(&now));

	/* Use program name as prefix */
	p = outputbuf;
	p += snprintf(p, LOG_OUTPUTBUF_LENGTH, "%s: ", _prgname);

	/* If the log level has been set to LOG_DEBUG, prefix all log messages with source
	   file name, function name and source file line number */
	if (_loglevel == LOG_DEBUG)
	{
		p += snprintf(p, LOG_OUTPUTBUF_LENGTH - (p-outputbuf), "(%s : %s : %d) ",
		              file, func, line);
	}

	/* Format log message */
	va_start(va, fmt);
	p += vsnprintf(p, LOG_OUTPUTBUF_LENGTH - (p-outputbuf), fmt, va);
	va_end(va);

	/* Make sure there's a newline at the end */
	if (*(p-1) != '\n')
	{
		*p = '\n';
		*(p+1) = '\0';
	}

	/* Split up output into single lines */
	for (linep = p = outputbuf; *p != '\0'; p++)
	{
		if (*p == '\n')
		{
			/* Write line without prefix to stdout/stderr, if we were told to do so */
			if (errmsg)
			{
				fputs(linep, stderr);
				fflush(stderr);
			}

			/* Write line to logfile */
			*p = '\0';
			fprintf(_logfd, "%s %s\n", timestampbuf, linep);

			/* Flush logfile, if the message is important enough */
			if (severity <= LOG_WARNING)
				fflush(_logfd);

			/* Advance to next line */
			linep = p++;
		}
	}
}
