/*
** $Id: main.c,v 1.11 2003/01/10 01:01:17 pief Exp $
**
** makelist - MetalServe file list generator
** Copyright (c) 2002-2003 by Pieter Hollants <contact@metalserve.org>
**
** This program is licensed under the GNU General Public License, version 2,
** as published by the Free Software Foundation and available in the file
** COPYING and the Internet location http://www.gnu.org/licenses/gpl.html.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
** or FITNESS FOR A PARTICULAR PURPOSE.
**
** Main program
**
** $Log: main.c,v $
** Revision 1.11  2003/01/10 01:01:17  pief
** Adapted makelist to changes to configuration parser and #defines.
** Added check so that loglevel can't be arbitrarily increased.
** Added print_config() function.
** Extended copyright note to the new year.
**
** Revision 1.10  2002/12/01 23:37:03  pief
** Removed no longer-present "flags" definition for CFGDIRECTIVE structures
**
** Revision 1.9  2002/11/24 11:05:39  pief
** Added configuration directive and command line switch to control
** verboseness.
** Added support for "loglevel" configuration directive.
** Adapted definition of configuration directives to conf.h changes.
**
** Revision 1.8  2002/10/25 16:08:31  pief
** Modified program banner
** Added logging support
**
** Revision 1.7  2002/10/19 22:35:25  pief
** Renamed makelist.[ch] to main.[ch]
**
** Revision 1.6  2002/10/19 20:51:38  pief
** Added missing #include's to get makelist compiling under FreeBSD
** Added support for the PACKAGE_NAME and PACKAGE_VERSION defines set by the
** top-level configure script
**
** Revision 1.5  2002/10/01 07:54:54  pief
** Moved rotation delay into #define in makelist.h
**
** Revision 1.4  2002/10/01 07:52:46  pief
** Fixed bug in insert_scannode() that caused directories to be unsorted
** when inserted in a certain order that never happened in reality (mentioned
** in metalserve.org news on 20092002, forgot to reactivate the fix)
**
** Revision 1.3  2002/09/30 22:19:40  pief
** Fixes due to wrong CVS commit
**
** Revision 1.2  2002/09/30 22:03:47  pief
** Implemented a progress display with a rotating cursor, which is displayed
** only if a tty is attached, and if so, also respects the terminal's width.
** Made metalserve more verbose in general.
** Added -q command line switch to make metalserve shut up again =)
** Moved code to convert a pure seconds value to minutes and seconds to a
** new common/misc.c routine named get_mins_and_secs()
**
** Revision 1.1  2002/09/24 21:46:03  pief
** Added first version of makelist main files
**
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include <errno.h>
#include <time.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <libgen.h>

#include "../lib/getopt.h"

#include "../common/base.h"
#include "../common/list.h"
#include "../common/conf.h"
#include "../common/log.h"
#include "../common/misc.h"

#include "main.h"
#include "mpegaudio.h"
#include "waveaudio.h"

const char *_prgbanner =
	"%s [%s v%s] - %s file list generator\n"
	"Copyright (c) 2002-2003 by Pieter Hollants <contact@metalserve.org> - http://www.metalserve.org\n\n";

char *_prgname;

BOOL _call_print_config = FALSE;

BOOL _tty_attached = FALSE;
int _terminal_width;

const char *_rotation_chars = "|/-\\";
#define ROTATION_CHARS_LEN 4
int _rotation = 0;
static struct timeval _last_rotation, _now_tv;

/*
** Command line options
*/
static struct option long_options[] =
{
	{ "configfile",		required_argument,	0, 'f' },
	{ "help",		no_argument,		0, 'h' },
	{ "quiet",		no_argument,		0, 'q' },
	{ "serverroot",		required_argument,	0, 's' },
	{ "test-config",	no_argument,		0, 't' },
	{ "verbose",		no_argument,		0, 'v' },
	{ "version",		no_argument,		0, 'V' },
	{ 0, 0, 0, 0 }
};
const char *_cmdlineopts = "fhqstvV";

const char *_cmdlinehelp =
	"This program is licensed under the GNU General Public License, version 2.\n"
	"See the file COPYING or visit http://www.gnu.org/licenses/gpl.html for details.\n\n"

	"Usage: %s [options]\n\n"

	"Options:\n"
	"  -f, --configfile=FILE           specify alternative configuration file\n"
	"  -h, --help                      list available command line options (this text)\n"
	"  -q, --quiet                     suppress output (for use when started by a cron job)\n"
	"  -s, --serverroot=PATH           override configured serverroot (root directory to descend)\n"
	"  -t, --test-config               test configuration file\n"
	"  -v, --verbose                   increase verboseness (loglevel)\n"
	"  -V, --version                   print version and exit\n";

static BOOL _quiet = FALSE;
static int _loglevel = DEF_LOGLEVEL;

/*
** Configuration file variables and directives
*/
char *_cfgfile_name;

char *_logfile_name;
char *_serverroot;
char *_compress_command;
BOOL *_keep_uncompressed_list;
char **_include_extensions;
char *_list_header;
BOOL *_dir_overview;
char **_nickname;

/* Valid arguments and associated values to the "loglevel" configuration directive. We intentionally
   don't try to get the strings from syslog.h, since this is what the user specifies in the config
   file and that should be under our control. */
CFGSTRVALUE _loglevelargs[] =
{
	/* string,	value,		exclusive? */
	{ "critical",	LOG_CRIT,	TRUE },
	{ "error",	LOG_ERR,	TRUE },
	{ "warning",	LOG_WARNING,	TRUE },
	{ "notice",	LOG_NOTICE,	TRUE },
	{ "info",	LOG_INFO,	TRUE },
	{ "debug",	LOG_DEBUG,	TRUE },
	LAST_CFGSTRVALUE
};

/* Configuration directives (see src/common/conf.h) */
CFGDIRECTIVE _directives[] =
{
	/* directive name,              argparser function,                     target variable,
	   min, max,                    section,                                params */
	{ "logfile",			argparser_string,			&_logfile_name,
	   0, FILENAME_MAX,		GLOBAL_SECTION,				NULL },
	{ "loglevel",			argparser_stringlist,			&_loglevel,
	   0, 1,			GLOBAL_SECTION,				_loglevelargs },
	{ "serverroot",			argparser_string,			&_serverroot,
	   0, PATH_MAX,			GLOBAL_SECTION,				NULL },
	{ "compress_command",		argparser_string,			&_compress_command,
	   0, FILENAME_MAX,		GLOBAL_SECTION,				NULL },
	{ "keep_uncompressed_list",	argparser_switch,			&_keep_uncompressed_list,
	   0, 0,			GLOBAL_SECTION,				NULL },
	{ "include_extensions",		argparser_stringlist,			&_include_extensions,
	   0, MAX_VALIST_PTRS,		GLOBAL_SECTION,				NULL },
	{ "list_header",		argparser_string,			&_list_header,
	   0, FILENAME_MAX,		GLOBAL_SECTION,				NULL },
	{ "dir_overview",		argparser_switch,			&_dir_overview,
	   0, 0,			GLOBAL_SECTION,				NULL },
	{ "nickname",			argparser_stringlist,			&_nickname,
	   0, MAX_VALIST_PTRS,		GLOBAL_SECTION,				NULL },
	LAST_CFGDIRECTIVE
};

/*
** Filelist variables
*/
LIST *_filelist;
time_t _starttime, _now;
int _mins_passed, _secs_passed;
int _totalkbytes, _totalfiles;

char _filelistname[NAME_MAX];
FILE *_fd = NULL;

unsigned char _buf[BUF_SIZE];
int _pad;

/*
** rc = parse_cmdline_args(argc, argv)
**
** Parses the program's command line arguments.
**
** Returns TRUE if the program should continue to run or FALSE if the
** program should quit.
*/
static BOOL parse_cmdline_args(int argc, char **argv)
{
	int c, option_index = 0;

	while (TRUE)
	{
		c = getopt_long(argc, argv, _cmdlineopts, long_options, &option_index);
		if (c == -1) break;

		switch (c)
		{
			case 'f':
			{
				_cfgfile_name = strdup(optarg);
				break;
			}
			case 'h':
			{
				printf(_prgbanner, _prgname, PACKAGE_NAME, PACKAGE_VERSION, PACKAGE_NAME);
				printf(_cmdlinehelp, _prgname);
				return FALSE;
			}
			case 'q':
			{
				_quiet = TRUE;
				break;	
			}
			case 's':
			{
				_serverroot = strdup(optarg);
				break;
			}
			case 't':
			{
				_call_print_config = TRUE;
				break;
			}
			case 'v':
			{
				_loglevel++;
				break;
			}
			case 'V':
			{
				printf(_prgbanner, _prgname, PACKAGE_NAME, PACKAGE_VERSION, PACKAGE_NAME);
				printf("Compiled %s %s\n", __DATE__, __TIME__);
				return FALSE;
			}
			case '?':
			{
				/* getopt_long already prints a usage message */
				return FALSE;
			}
		}
	}

	return TRUE;
}

/*
** rc = parse_cfgfile()
**
** Parses the MetalServe configuration file.
**
** Returns TRUE if the program should continue to run or FALSE if the
** program should quit.
*/
static BOOL parse_cfgfile(void)
{
	int attempt;
	CFGFILE *cfgfile = NULL;
	char *errstr;
	char *old_serverroot;
	int old_loglevel;

	/* Attempt to open configuration file at one of two locations */
	for (attempt = 0; attempt < 2; attempt++)
	{
		/* Try the system-wide location first */
		if (attempt == 0)
			_cfgfile_name = DEF_CFGFILE_NAME;

		/* Then try the home dir of the user we're running under */
		else
		{
			char *homedir;

			/* Get path of current user's home directory and allocate memory for composed filename */
			homedir = getenv("HOME");
			_cfgfile_name = malloc(PATH_MAX);
			if ( (!homedir) || (!_cfgfile_name) )
			{
				fprintf(stderr, "%s: could not compose full filename of \"%s\" configuration file!\n",
					_prgname, HOMEDIR_CFGFILE_NAME);
				return FALSE;
			}
			
			snprintf(_cfgfile_name, PATH_MAX, "%s/%s", homedir, HOMEDIR_CFGFILE_NAME);
		}

		/* Now try to actually open the configuration file */
		errno = 0;
		cfgfile = cfgopen(_cfgfile_name, CFGOPENFLAGS_IGNUNKNOWN);
		if ( (cfgfile) || ( (!cfgfile) && (errno != ENOENT) ) )
			break;
	}
	
	/* Couldn't we open it for some reason? */
	if (!cfgfile)
	{
		/* File did not exist at both locations */
		if (errno == ENOENT)
		{
			fprintf(stderr, "%s: missing configuration file at either \"%s\" or \"%s\"!\n",
				_prgname, DEF_CFGFILE_NAME, HOMEDIR_CFGFILE_NAME);
		}
		/* Some other serious error */
		else
		{	
			fprintf(stderr, "%s: error opening configuration file \"%s\":\n%s!\n",
				_prgname, _cfgfile_name, strerror(errno));
		}

		return FALSE;
	}

	/* Process configuration directives (but preserve variables eventually modified on command line) */
	old_serverroot = _serverroot;
	old_loglevel = _loglevel;
	_loglevel = 0;
	if (!cfgread(cfgfile, _directives))
	{
		if (cfgerrno(cfgfile)) errstr = cfgerrstr(cfgfile);
		else errstr = strerror(errno);
		fprintf(stderr, "%s: error while reading line %d of \"%s\":\n%s!\n",
			_prgname, cfglineno(cfgfile), _cfgfile_name, errstr);
		return FALSE;
	}
	if (old_serverroot) _serverroot = old_serverroot;
	if (_loglevel)
	{
		_loglevel += (old_loglevel - DEF_LOGLEVEL);
		if (_loglevel > LOG_DEBUG)
			_loglevel = LOG_DEBUG;
	}
	else
		_loglevel = old_loglevel;

	cfgclose(cfgfile);

	return TRUE;
}

/*
** print_config()
**
** Dumps the current configuration to stdout. Invoked when makelist is
** started with the "-t" (test configfile) argument.
*/
static void print_config(void)
{
	char **strlistentry;

        printf("%s configuration dump:\n\n", _prgname);

        printf("Directive                Value                     Description\n");
        printf("---------------------------------------------------------------------------------------------------\n");

        printf("logfile                  %-24.24s  Destination for log messages\n",
               (_logfile_name ? _logfile_name : "(disabled)"));
        printf("loglevel                 %-24.24s  Controls the amount of messages logged\n\n",
               getloglevelname(_loglevel));
        
        printf("serverroot               %-24.24s  Root directory from which files are served\n\n",
               _serverroot);

	if (_compress_command)
		printf("compress_command         %-24.24s  Command to compress the generated filelist\n",
		       _compress_command);
	else
		printf("compress_command         %-24.24s  Command to compress the generated filelist\n", "(None)");
	printf("keep_uncompressed_list   %-24.24s  Keep the uncompressed list after it was compressed\n",
	       (_keep_uncompressed_list ? "Yes"  : "No"));
	if (_include_extensions)
	{
	        for (strlistentry = _include_extensions; *strlistentry; strlistentry++)
		{
			if (strlistentry == _include_extensions)
				printf("include_extensions       %-24.24s  File extensions of files to be included\n",
				       *strlistentry);
			else
				printf("                         %s\n", *strlistentry);
		}
	}
	else
		printf("include_extensions       %-24.24s  File extensions of files to be included", "*");

	if (_list_header)
		printf("list_header              %-24.24s  File to include as header in filelist\n",
		       _list_header);
	else
		printf("list_header              %-24.24s  File to include as header in filelist\n", "(None)");

	printf("dir_overview             %-24.24s  Include directory overview in filelist\n\n",
	       (_dir_overview ? "Yes" : "No"));

	printf("nickname                 %-24.24s  Nickname used in the filelist (only the first is used)\n\n",
	       *_nickname);

        printf("Note: this dump includes only directives relevant to %s. Call metalserved\n"
               "with the -t argument to see the values of additional directives.\n", _prgname);
}

/*
** width = determine_terminal_width()
**
** Determine the width of the attached terminal.
*/
static int determine_terminal_width(void)
{
	struct winsize wsz;

	if (ioctl(STDOUT_FILENO, TIOCGWINSZ, &wsz) < 0)
	{
		fprintf(stderr, "%s: could not determine terminal width:\n%s!\n",
			_prgname, strerror(errno));
		return 0;
	}

	return wsz.ws_col;
}

/*
** rc = start_filelist()
**
** Opens the filelist and writes the MetalServe header.
*/
static BOOL start_filelist(void)
{
	/* Initialize counters */
	_totalkbytes = 0;
	_totalfiles = 0;

	/* Open file list for writing */
	_fd = fopen(_filelistname, "w");
	if (!_fd)
	{
		fprintf(stderr, "%s: error opening file \"%s\" for writing:\n%s!\n",
			_prgname, _filelistname, strerror(errno));
		return FALSE;
	}

	/* Write MetalServe header */
	fprintf(_fd, "\n");
	_pad = (83 - strlen(*_nickname)) / 2;
	fprintf(_fd, "%*c-{{ List of %s's shared files }}-\n", _pad, ' ', *_nickname);
	fprintf(_fd, "\n");

	strftime(_buf, sizeof(_buf), "%a %d-%B-%Y %T GMT%z", localtime(&_starttime));
	_pad = (83 - strlen(_buf) - strlen(PACKAGE_VERSION)) / 2;
	fprintf(_fd, "%*cLast updated on %s by MetalServe %s\n", _pad, ' ', _buf, PACKAGE_VERSION);
	fprintf(_fd, "%*chttp://www.metalserve.org\n", 45, ' ');

	fprintf(_fd, "\n");
	fprintf(_fd, "-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n");
	fprintf(_fd, "\n");

	/* Include auxiliary list header, if configured */
	if (_list_header)
	{
		FILE *headerfd;
		
		/* Attempt to open header */
		headerfd = fopen(_list_header, "r");
		if (headerfd)
		{
			/* Copy header, line by line */
			while (fgets(_buf, sizeof(_buf), headerfd))
				fputs(_buf,_fd);
			if (ferror(headerfd))
			{
				fprintf(stderr, "%s: error while reading from list_header file \"%s\":\n%s!\n",
					_prgname, _list_header, strerror(errno));
				return FALSE;					
			}

			/* Print additional seperator for instructions below */
			fprintf(_fd, "\n");
			fprintf(_fd, "-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n");
			fprintf(_fd, "\n");
		}
		else
		{
			fprintf(stderr, "%s: error opening list_header file \"%s\" for reading:\n%s!\n",
				_prgname, _list_header, strerror(errno));
			return FALSE;		
		}
	}

	/* Write instructions header */
	fprintf(_fd, "MetalServe instructions:\n\n");

	fprintf(_fd, "To receive a file from me, type \"!%s <exact filename>\" in a channel window\n", *_nickname);
	fprintf(_fd, "or copy and paste one of the lines below into a channel I'm in.\n\n");

	fprintf(_fd, "For example: !%s Metal rules.mp3\n\n", *_nickname);

	fprintf(_fd, "Type @%s-help to receive additional help.\n\n", *_nickname);

	return TRUE;
}

/*
** rc = dirent_selector(dirent)
**
** Callback function called by glibc's scandir() function. Examines the passed
** dirent structure and returns zero for entries that should be excluded.
*/
static int dirent_selector(const struct dirent *entry)
{
	/* Exclude "." and ".." entries */
	if (entry->d_name[0] == '.')
	{
		if (entry->d_name[1] == '\0') return 0;
		if ( (entry->d_name[1] == '.') && (entry->d_name[2] == '\0') ) return 0;
	}

	/* We can't put the include_extensions code here since entry->d_type
	   is not guaranteed to be available in a usable manner on all platforms. */

	return 1;
}

/*
** insert_scannode(list, node)
**
** Inserts the specified SCANNODE at the "correct" position in "list". The
** list is sorted so that directories come alphabetically first, followed
** by the files in alphabetical order.
*/
static void insert_scannode(LIST *list, SCANNODE *node)
{
	SCANNODE *current;
	BOOL zero_ok = FALSE;

	for (current = (SCANNODE *)list->first; current; current = (SCANNODE *)current->next)
	{
		/* If the node to be inserted is a directory, stop search as soon
		   as the first file is reached */
		if ( (IS_DIR(node)) && (!(IS_DIR(current))) )
		{
			current = (SCANNODE *)current->prev;
			zero_ok = TRUE;
			break;
		}

		/* Likewise if the node to be inserted is a file, skip the directory
		   nodes at the beginning of the list */
		if ( !(IS_DIR(node)) && (IS_DIR(current)) )
			continue;

		/* Find the alphabetically correct place */
		if (strcasecmp(node->name, current->name)<0)
		{
			current = (SCANNODE *)current->prev;
			zero_ok = TRUE;
			break;
		}
	}

	/* Should we have passed the last node, we need to go back for it for
	   insert_node() to work properly */
	if ((!current) && (!zero_ok)) current = (SCANNODE *)list->last;

	/* Aha, we found a suitable place */
	insert_node(list, (NODE *)current, (NODE *)node);
}

/*
** rc = process_directory(list, dirname)
**
** Processes the specified directory and adds SCANNODEs for matching files and
** directories to "list". Skips those files that don't match the configured
** include extensions.
*/
static BOOL process_directory(LIST *list, char *dirname)
{
	struct dirent **entries;
	int i, num_entries;

	/* Scan directory */
	num_entries = scandir(dirname, &entries, dirent_selector, alphasort);
	if (num_entries < 0)
	{
		logmsg_stderr(LOG_ERR, "error while scanning directory \"%s\":\n%s!\n",
		              dirname, strerror(errno));
		return FALSE;
	}

	for (i = 0; i < num_entries; i++)
	{
		char filename[NAME_MAX];
		struct stat statbuf;
		SCANNODE *node;

		/* If a tty is attached, update rotating cursor four times each second */
		if (_tty_attached)
		{
			gettimeofday(&_now_tv, NULL);
			if ((unsigned long)_now_tv.tv_usec - (unsigned long)_last_rotation.tv_usec > ROTATION_DELAY)
			{
				printf("\r%c\r", _rotation_chars[_rotation]);
				fflush(stdout);

				_last_rotation.tv_usec = _now_tv.tv_usec;
				if (_rotation < ROTATION_CHARS_LEN - 1)
					_rotation++;
				else _rotation = 0;
			}
		}

		/* Examine entry type and file size */
		snprintf(filename, NAME_MAX-1, "%s/%s", dirname, entries[i]->d_name);
		if (stat(filename, &statbuf) < 0)
		{
			logmsg_stderr(LOG_ERR, "error stat()'ing file \"%s\":\n%s!\n",
			              filename, strerror(errno));
			return FALSE;
		}

		/* Was an "include_extensions" configuration directive given and is this a directory? */
		if ( !(S_ISDIR(statbuf.st_mode)) && (_include_extensions) )
		{
			char *current_extension, **extension;
			BOOL match = FALSE;

			/* If so, determine the current file's extension... */
			current_extension = strrchr(entries[i]->d_name, '.');
			if (!current_extension)
				continue;

			/* ...and compare with the configured file extensions */
			for (extension = _include_extensions; (!match) && (*extension); extension++)
			{
				if (!strcasecmp(current_extension, *extension))
					match = TRUE;
			}
			if (!match) continue;
		}

		/* Allocate SCANNODE structure for this entry */
		node = (SCANNODE *)calloc(sizeof(SCANNODE), 1);
		if (!node)
		{
			logmsg_stderr(LOG_ERR, "out of memory while processing directory entry \"%s\"!\n",
				      filename);
			return FALSE;
		}
		node->name = strdup(entries[i]->d_name);
		node->fullname = strdup(filename);

		if (S_ISDIR(statbuf.st_mode))
		{
			/* If a tty is attached, print progress line for this subdirectory */
			if (_tty_attached)
			{
				printf("\r%c %-*.*s\r", _rotation_chars[_rotation],
				       _terminal_width - 3, _terminal_width - 3, filename);
				fflush(stdout);
			}

			/* Allocate list */
			node->contents = allocate_list();
			if (!node->contents)
			{
				logmsg_stderr(LOG_ERR, "out of memory while processing directory \"%s\"!\n",
				              filename);
				return FALSE;
			}

			insert_scannode(list, node);

			/* Recursively call ourself for the contents of the directory */
			if (!process_directory(node->contents, filename))
				return FALSE;
		}
		else
		{
			FILE *file;
			char *extension;

			/* Examine file */
			node->size = statbuf.st_size;

			file = fopen(filename, "r");
			if (!file)
			{
				node->infotype = INFOTYPE_NONE;
				insert_scannode(list, node);
				continue;
			}
			
			/* Filetype detection goes here */
			extension = strrchr(node->name, '.');
			if (!(strncasecmp(extension, ".wav", 4)))
				examine_waveaudio_file(file, filename, node);
			else
				examine_mpegaudio_file(file, filename, node);

			/* Close file and insert node into filelist */
			fclose(file);
			insert_scannode(list, node);
		}
	}

	return TRUE;
}

/*
** write_dir_overview_entry(startlist);
**
** Subfunction to write_dir_overview() below.
*/
static void write_dir_overview_entry(LIST *startlist)
{
	SCANNODE *current;

	/* Print lines for subdirectories */
	for (current = (SCANNODE *)startlist->first; current; current = (SCANNODE *)current->next)
	{
		if (current->contents)
		{
			/* Print line for current directory */
			fprintf(_fd, "%s\n", current->fullname + strlen(_serverroot));

			/* Process subdirectories as well */
			write_dir_overview_entry(current->contents);
		}
	}
}

/*
** write_dir_overview(startlist)
** 
** Writes an overview of the scanned directories in front of the
** actual filelist.
*/
static void write_dir_overview(LIST *startlist)
{
	fprintf(_fd, "-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n");
	fprintf(_fd, "\n");
	
	fprintf(_fd, "Overview of the directories in this filelist (actual file listing follows):\n\n");

	write_dir_overview_entry(startlist);	

	fprintf(_fd, "\n");
	fprintf(_fd, "-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n");
	fprintf(_fd, "\n");
}

/*
** write_filelist(startlist, name)
**
** Parses the filelist tree and generates entries for the directories and
** files found. "startlist" is the LIST structure of SCANNODEs to be parsed
** and "name" is an associated name. This is a recursive function.
*/
static void write_filelist(LIST *startlist, char *name)
{
	SCANNODE *current;
	BOOL header_written = FALSE;

	/* Process subdirectories first */
	for (current = (SCANNODE *)startlist->first; current; current = (SCANNODE *)current->next)
	{
		if (current->contents)
		{
			/* Write virtual path relative to _serverroot instead of actual filesystem path */
			write_filelist(current->contents, current->fullname + strlen(_serverroot));
		}
	}

	/* Then process the files */
	for (current = (SCANNODE *)startlist->first; current; current = (SCANNODE *)current->next)
	{
		/* We handled subdirectories above, now deal with the current dir's files */
		if (!(IS_DIR(current)))
		{
			/* Write header line for the current directory, if not already done */
			if (!header_written)
			{
				fprintf(_fd, "-----[ %s ]-----\n\n", name);
				header_written = TRUE;
			}
			
			/* Write nickname and filename left-bound */
			sprintf(_buf, "!%s %s", *_nickname, current->name);
			fprintf(_fd, "%-74.74s ::INFO:: ", _buf);

			/* Append file size in human-readable notation */
			if (current->size > 1024*1024)
				sprintf(_buf, "%4.1fM", ((float)current->size / (1024*1024)) );
			else sprintf(_buf, "%4dK", (int)current->size / 1024);
			fprintf(_fd, "%s", _buf);

			/* Update filelist statistics */
			_totalkbytes += (current->size / 1024);
			_totalfiles++;

			/* Append filetype-specific information, if available */
			switch (current->infotype)
			{
				case INFOTYPE_MPEGAUDIO:
					get_mpegaudio_info_string((MPEGAUDIOINFO *)current->info, _buf);
					fprintf(_fd, " %s\n", _buf);
					break;

				case INFOTYPE_WAVEAUDIO:
					get_waveaudio_info_string((WAVEAUDIOINFO *)current->info, _buf);
					fprintf(_fd, " %s\n", _buf);
					break;

				default:
				{
					fprintf(_fd, "\n");
				}
			}
		}
	}

	/* If a header was written for this directory, we also wrote lines for at
	   least one file, so we add another linefeed to optically signalize the end
	   of this directory. */
	if (header_written) fprintf(_fd, "\n");
}

/*
** rc = finish_filelist()
**
** Finishes the filelist with a footer with statistics.
*/
static BOOL finish_filelist(void)
{
	float totalsize;

	errno = 0;

	fprintf(_fd, "-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n");
	fprintf(_fd, "\n");

	/* Calculate how long we took */
	get_mins_and_secs(time(NULL) - _starttime, &_mins_passed, &_secs_passed);

	totalsize = ((float)_totalkbytes) / (1024 * 1024);
	if (!totalsize)
	{
		totalsize = ((float)_totalkbytes) / 1024;
		fprintf(_fd, "List totals: %2.2f MB in %d file(s)\n", totalsize, _totalfiles);
	}
	else fprintf(_fd, "List totals: %2.2f GB in %d file(s)\n", totalsize, _totalfiles);
	fprintf(_fd, "List generation took %d minute(s) and %d second(s)\n\n", _mins_passed, _secs_passed);
	fprintf(_fd, "Generated by makelist (MetalServe %s) - http://www.metalserve.org/\n", PACKAGE_VERSION);

	fclose(_fd);
	_fd = NULL;

	if (errno) return FALSE;
	else return TRUE;
}

/*
** rc = compress_filelist()
**
** Processes the configured compress_command and launches it to compress the generated
** filelist. Replaces the placeholders %a with the filename for the compressed archive (without
** suffix) and %u with the filename of the uncompressed list. If compression was successful,
** the uncompressed list is deleted unless the "keep_uncompressed_list" directive was set to
** "true" or "yes".
*/
static BOOL compress_filelist(void)
{
	char archive_name[FILENAME_MAX], *tmp;
	char *p, *q;
	int argslen, pid, status;

	/* Compose archive filename (without suffix) */
	strftime(_buf, sizeof(_buf), "%m-%d-%Y", localtime(&_starttime));
	sprintf(archive_name, "%s-[%s]", *_nickname, _buf);

	/* Generate copy of compress command arguments with replaced placeholders */
	argslen = strlen(_compress_command) + 2 * FILENAME_MAX;
	tmp = malloc(argslen);
	if (!tmp)
	{
		logmsg_stderr(LOG_ERR, "not enough memory to process \"compress_command\"!\n");
		return FALSE;
	}

	p = _compress_command;
	q = tmp;
	while ((*p) && ((q - tmp) < argslen))
	{
		if ( (*p == '%') && (*(p+1) == 'a') )
		{
			strncpy(q, archive_name, FILENAME_MAX);
			p += 2;
			q += strlen(archive_name);
		}
		else if ( (*p == '%') && (*(p+1) == 'u') )
		{
			strncpy(q, _filelistname, FILENAME_MAX);
			p += 2;
			q += strlen(_filelistname);
		}
		else *q++ = *p++;
	}
	free(_compress_command);
	_compress_command = tmp;

	/* Instead of the interrupt-resistent system() call, we execute the
	   compression program as a child process. */
	pid = fork();

	/* Did something go wrong? */
	if (pid == -1)
	{
		logmsg_stderr(LOG_ERR, "could not fork() to launch \"compress_command\":\n%s!\n",
		              strerror(errno));
		return FALSE;
	}

	/* Are we the child? */
	if (pid == 0)
	{
		char *argv[4];

		/* Redirect in- and output to /dev/null */
		freopen("/dev/null", "r", stdin);
		freopen("/dev/null", "w", stdout);
		freopen("/dev/null", "w", stderr);

		/* Construct argument array for execve() call */
		argv[0] = "/bin/sh";
		argv[1] = "-c";
		argv[2] = _compress_command;
		argv[3] = 0;

		/* execve() only returns if errors occured, and in that case with -1 */
		if (execve("/bin/sh", argv, NULL) == -1)
		{
			logmsg_stderr(LOG_ERR, "execution of compress_command \"%s\" failed:\n%s!\n",
		                      _compress_command, strerror(errno));
			return FALSE;
		}
	}

	/* Wait for child process to exit */
	while (TRUE)
	{
		if (waitpid(pid, &status, 0) == -1)
		{
			if (errno != EINTR)
				return FALSE;
		}
		else
		{
			/* Did the command return an error? */
			if (WEXITSTATUS(status))
			{
				logmsg_stderr(LOG_ERR, "execution of compress_command \"%s\" failed with /bin/sh return code %d!\n",
			                      _compress_command, WEXITSTATUS(status));
				return FALSE;
			}

			/* Delete original text file unless we were told not to do so */
			if (!_keep_uncompressed_list)
			{
				if (remove(_filelistname))
				{
					logmsg_stderr(LOG_ERR, "could not remove uncompressed textfile \"%s\":\n%s!\n",
				                      _filelistname, strerror(errno));
					return FALSE;
				}
			}

			return TRUE;
		}
	}
}

/*
** Main routine
*/
int main(int argc, char **argv)
{
	BOOL rc;
	char **extension;

	/* Determine program name */
	_prgname = (char *)basename(strdup(argv[0]));

	/* Parse command line arguments */
	if (!parse_cmdline_args(argc, argv))
		return 1;

	/* Parse configuration file */
	if (!parse_cfgfile())
		return 1;

	/* Check whether the -t option was given */
	if (_call_print_config)
	{
		printf(_prgbanner, _prgname, PACKAGE_NAME, PACKAGE_VERSION, PACKAGE_NAME);
		print_config();
		return 0;
	}

	if (!_quiet)
	{
		/* Print program banner */
		printf(_prgbanner, _prgname, PACKAGE_NAME, PACKAGE_VERSION, PACKAGE_NAME);

		/* Determine whether a terminal is attached */
		if (isatty(STDOUT_FILENO))
		{
			/* Determine terminal width */
			_terminal_width = determine_terminal_width();
			if (!_terminal_width)
				return 1;

			_tty_attached = TRUE;
		}
	}

	/* Allocate memory for the filelist */
	_filelist = allocate_list();
	if (!_filelist)
	{
		fprintf(stderr, "%s: not enough memory for filelist!\n",
		        _prgname);
		return FALSE;
	}

	/* The special argument "none" disables logging at all */
	if (!strncmp(_logfile_name, "none", 4))
		_logfile_name = NULL;
		
	/* Check whether a logfile has been specified */
	if (_logfile_name)
	{
		/* If so, open it */
		if (logopen(_logfile_name, _loglevel))
		{
			logmsg(LOG_NOTICE, "%s [%s v%s] starting\n", _prgname,
			       PACKAGE_NAME, PACKAGE_VERSION);
		}
		else
		{
			fprintf(stderr, "%s: could not open logfile \"%s\" for writing:\n%s!\n",
				_prgname, _logfile_name, strerror(errno));
			return FALSE;
		}
	}

	/* Remember starttime for statistical purposes */
	_starttime = time(NULL);

	/* Process the directory tree and add files to the list */
	if (!_quiet)
		printf("Scanning \"%s\" ...\n", _serverroot);
	logmsg(LOG_INFO, "scanning serverroot \"%s\" ...\n", _serverroot);

	rc = process_directory(_filelist, _serverroot);

	/* Start the on-disk filelist */
	if (rc)
	{
		/* Generate file list name */
		strftime(_buf, sizeof(_buf), "%m-%d-%Y", localtime(&_starttime));
		sprintf(_filelistname, "%s-[%s].txt", *_nickname, _buf);

		if (!_quiet)
		{
			snprintf(_buf, BUF_SIZE, "Writing filelist \"%s\" ...", _filelistname);
			printf("\r%-*s\n", _terminal_width, _buf);
			fflush(stdout);
		}
		logmsg(LOG_INFO, "writing filelist \"%s\" ...\n", _filelistname);

		rc = start_filelist();
	}

	/* Write directory overview, if enabled */
	if ( (rc) && (_dir_overview) )
		write_dir_overview(_filelist);

	/* Write files from LIST structure to on-disk filelist */
	if (rc)
		write_filelist(_filelist, "/");

	/* Finish on-disk filelist */
	if (rc)
		rc = finish_filelist();

	/* Compress filelist, if enabled */
	if ( (rc) && (_compress_command) )
	{
		if (!_quiet)
		{
			printf("Compressing filelist ...\n");
			fflush(stdout);
		}
		logmsg(LOG_INFO, "compressing filelist ...\n");

		if (!compress_filelist())
			return 1;
	}

	/* If errors occured, attempt to delete the incomplete textfile */
	if (!rc)
	{
		if (_fd)
			fclose(_fd);
		unlink(_filelistname);	
	
		return 1;
	}

	/* Print statistics */
	if (!_quiet)
		printf("\nTotal time elapsed: %02d:%02d\n", _mins_passed, _secs_passed);
	logmsg(LOG_NOTICE, "filelist generation took %02d:%02d\n", _mins_passed, _secs_passed);

	return 0;
}
