/*
** $Id: mpegaudio.c,v 1.3 2003/01/10 00:59:34 pief Exp $
**
** makelist - MetalServe file list generator
** Copyright (c) 2002-2003 by Pieter Hollants <contact@metalserve.org>
**
** This program is licensed under the GNU General Public License, version 2,
** as published by the Free Software Foundation and available in the file
** COPYING and the Internet location http://www.gnu.org/licenses/gpl.html.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
** or FITNESS FOR A PARTICULAR PURPOSE.
**
** MPEG Audio information functions
**
** $Log: mpegaudio.c,v $
** Revision 1.3  2003/01/10 00:59:34  pief
** Extended copyright note to the new year
**
** Revision 1.2  2002/10/19 22:35:25  pief
** Renamed makelist.[ch] to main.[ch]
**
** Revision 1.1  2002/09/24 21:46:23  pief
** Added first version of MPEG Audio-related functions
**
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include <errno.h>

#include "../common/base.h"
#include "../common/list.h"

#include "main.h"
#include "mpegaudio.h"

extern char *_prgname;

/* Bitrates as defined in the MPEG standard. Three-dimensional array accessed
   with MPEG version, MPEG layer and bitrate index (bitmask read from the file).
   Don't ask why they seem to be reserved, this is due to the way the meaning
   of the bits in the MPEG frame header is defined.*/
const int mpa_bitrates[][3][15] =
{
	/* MPEG-2 */
	{
		/* Layer 3 */
		{
			0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160
		},
		/* Layer 2 */
		{
			0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160
		},
		/* Layer 1 */
		{
			0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256
		}
	},
	/* MPEG-1 */
	{
		/* Layer 3 */
		{
			0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320
		},
		/* Layer 2 */
		{
			0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384
		},
		/* Layer 1 */
		{
			0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448
		}
	}
};

/* Sampling frequencies as defined in the MPEG standard. Two-dimensional array accessed
   with MPEG version and sampling frequency index (bitmask read from the file). As above,
   the order is due to the meaning of the bits in the MPEG frame header. */
const int mpa_samplingfreqs[][3] =
{
	/* MPEG v2 */
	{
		22050, 24000, 16000
	},
	/* MPEG v1 */
	{
		44100, 48000, 32000
	}
};

/* ID3v1 genres */
const char *id3v1_genres[] =
{
	"Blues", "Classic Rock", "Country", "Dance", "Disco", "Funk", "Grunge",
	"Hip-Hop", "Jazz", "Metal", "New Age", "Oldies", "Other", "Pop", "R&B",
	"Rap", "Reggae", "Rock", "Techno", "Industrial", "Alternative", "Ska",
	"Death Metal", "Pranks", "Soundtrack", "Euro-Techno", "Ambient",
	"Trip-Hop", "Vocal", "Jazz+Funk", "Fusion", "Trance", "Classical",
	"Instrumental", "Acid", "House", "Game", "Sound Clip", "Gospel",
	"Noise", "Alt", "Bass", "Soul", "Punk", "Space", "Meditative",
	"Instrumental Pop", "Instrumental Rock", "Ethnic", "Gothic", "Darkwave",
	"Techno-Industrial", "Electronic", "Pop-Folk", "Eurodance", "Dream",
	"Southern Rock", "Comedy", "Cult", "Gangsta Rap", "Top 40",
	"Christian Rap", "Pop/Funk", "Jungle", "Native American", "Cabaret",
	"New Wave", "Psychedelic", "Rave", "Showtunes", "Trailer", "Lo-Fi",
	"Tribal", "Acid Punk", "Acid Jazz", "Polka", "Retro", "Musical",
	"Rock & Roll", "Hard Rock", "Folk", "Folk/Rock", "National Folk",
	"Swing", "Fast-Fusion", "Bebob", "Latin", "Revival", "Celtic",
	"Bluegrass", "Avantgarde", "Gothic Rock", "Progressive Rock",
	"Psychedelic Rock", "Symphonic Rock", "Slow Rock", "Big Band",
	"Chorus", "Easy Listening", "Acoustic", "Humour", "Speech", "Chanson",
	"Opera", "Chamber Music", "Sonata", "Symphony", "Booty Bass",
	"Primus", "Porn Groove", "Satire", "Slow Jam", "Club", "Tango",
	"Samba", "Folklore", "Ballad", "Power Ballad", "Rhythmic Soul",
	"Freestyle", "Duet", "Punk Rock", "Drum Solo", "A Cappella",
	"Euro-House", "Dance Hall", "Goa", "Drum & Bass", "Club-House",
	"Hardcore", "Terror", "Indie", "BritPop", "Negerpunk", "Polsk Punk",
	"Beat", "Christian Gangsta Rap", "Heavy Metal", "Black Metal",
	"Crossover", "Contemporary Christian", "Christian Rock",
	"Merengue", "Salsa", "Thrash Metal", "Anime", "JPop", "Synthpop"
};

/*
** rc = check_mpegaudio_header(header, info)
**
** Checks whether the supplied 4-byte sequence "header" forms a valid MPEG
** audio frame header or not. The MPEGAUDIOINFO structure "info" will contain
** the extracted values and will also have been modified if the function
** returns FALSE. "header" is a "byte-concatenated" value as created using eg.
** the INT32_BE() MACRO in makelist.h.
**
** The interesting part of the header definition in [ISO11172-3]:
**
** AAAAAAAA AAABBCCD EEEEFFGH IIJJKLMM (32 bits / 4 bytes)
**
** A: Frame sync (11 bits all 1, resp. 12 bits for MPEG Audio v2.5)
** B: MPEG Audio Version ID (00 = v2.5, 01 = reserved, 10 = v2, 11 = v1)
** C: MPEG Audio Layer (00 = reserved, 01 = Layer III, 10 = Layer II, 11 = Layer I)
** [...]
** E: Bitrate index (depends on B and C, for MPEG-1 Layer III eg. 1001 = 128)
** F: Sampling rate frequency (depends on B, for MPEG-1 Layer III eg. 00 = 44100 Hz)
** G: Padding bit (0 = frame is not padded, 1 = frame is padded)
** [...]
** I: Channel mode (00 = stereo, 01 = Joint stereo, 10 = Dual mono, 11 = mono)
** [...]
**
** This function does not look for Xing/Frauenhofer tags etc.
**
** References:
** [ISO11172-3] "Coding of moving pictures and associated audio for digital
**              storage media at up to about 1.5 MBit/s - Part 3 (Audio)",
**              ISO document 11172-3, available at eg. www.mp3tech.org
*/
static BOOL check_mpegaudio_header(int header, MPEGAUDIOINFO *info)
{
	/* 11 bits frame sync */
	if ( ((header & 0xFFE00000) >> 21) != 0x7FF)
		return FALSE;

	/* Determine MPEG version */
	info->mpeg_version = (header & 0X180000) >> 19;
	if ( (info->mpeg_version < 2) || (info->mpeg_version > 3) )
		return FALSE;
	info->mpeg_version = info->mpeg_version - 2;

	/* Determine MPEG layer */
	info->mpeg_layer = (header & 0x60000) >> 17;
	if ( (info->mpeg_layer < 1) || (info->mpeg_layer > 3) )
		return FALSE;
	info->mpeg_layer--;

	/* Determine bitrate */
	info->bitrate = (header & 0xF000) >> 12;
	if ( (info->bitrate < 1) || (info->bitrate > 14) )
		return FALSE;
	info->bitrate = mpa_bitrates[info->mpeg_version][info->mpeg_layer][info->bitrate];

	/* Determine sampling frequency */
	info->samplingfreq = (header & 0xC00) >> 10;
	if (info->samplingfreq > 2)
		return FALSE;
	info->samplingfreq = mpa_samplingfreqs[info->mpeg_version][info->samplingfreq];

	/* Determine padding */
	if ( (header & 0x200) >> 9) info->padded = TRUE;
	else info->padded = FALSE;

	/* Determine channelmode */
	info->channelmode = (header & 0xC0) >> 6;
	if ( (info->channelmode < 0) || (info->channelmode > 3) )
		return FALSE;

	return TRUE;
}

/*
** rc = find_mpegaudio_header(file, info)
**
** Seeks in the specified opened file "file" for the next MPEG audio frame header
** from the file's current position onwards. "filename" should be the filename
** associated with "file". The MPEGAUDIOINFO structure "info" will contain the
** extracted values and will also have been modified if the function returns FALSE.
**
** The function returns FALSE if no header was found or errors occured (for
** which an error message will have been written to stderr), and TRUE if a header
** was found.
*/
static BOOL find_mpegaudio_header(FILE *file, char *filename, MPEGAUDIOINFO *info)
{
	unsigned char buf[MPEG_AUDIO_FRAME_HEADERLEN];
	unsigned int header;
	BOOL sync_found;

	/* Search for next MPEG audio frame header */
	memset(buf, 0, sizeof(buf));
	header = 0;
	for (sync_found = FALSE; !sync_found; (sync_found = check_mpegaudio_header(header, info)) )
	{
		int i;

		/* Shift previously read bytes (is assumed to be faster than seek n-times 3 bytes back) */
		for (i = 1; i < (sizeof(buf) + 1); i++) buf[i-1] = buf[i];

		/* Read another byte */
		if (!(fread(&buf[(sizeof(buf) - 1)], 1, 1, file)))
		{
			if (ferror(file))
			{
				fprintf(stderr, "%s: error while reading from file \"%s\": %s!\n",
					_prgname, filename, strerror(errno));
			}
			return FALSE;
		}

		/* Generate integer value for check_mpegaudio_header() function */
		header = INT32_BE(buf, 0);
	}
	if (!sync_found) return FALSE;

	return TRUE;
}

/*
** bytes_per_frame = calc_mpegaudio_framebodylength(bitrate)
**
** Given a MPEGAUDIOINFO structure with correctly filled mpeg_layer, samplingfreq
** and padded fields together with a bitrate, this function returns the length of
** an appropriate MPEG audio frame body (add 4 bytes to get the total length).
*/
static float calc_mpegaudio_framelength(MPEGAUDIOINFO *info, float bitrate)
{
	float bytes_per_frame;

	/* [ISO11172-3] page 30 gives the following formulas to calculate how
	   many bytes a frame uses (the first formula has already been adjusted
	   from slots to bytes - a slot is 4 bytes in MPEG Layer 1):

	  - for Layer 1:

	  bytes_per_frame = (12 * 4 * bitrate / sampling_frequency) + padding

	  - for Layers 2 and 3:

	  bytes_per_frame = (144 * bitrate / sampling_frequency) + padding */
	if (info->mpeg_layer == MPEG_AUDIO_LAYER1)
		bytes_per_frame = L1_BPF_MULT * bitrate;
	else
		bytes_per_frame = L2_L3_BPF_MULT * bitrate;
	bytes_per_frame /= (float)info->samplingfreq;
	if (info->padded) bytes_per_frame++;

	/* Subtract 4 bytes for the header */
	return (bytes_per_frame - MPEG_AUDIO_FRAME_HEADERLEN);
}

/*
** rc = examine_mpegaudio_file(FILE *file, char *filename, SCANNODE *node)
**
** Checks whether the passed file is a MPEG audio file and, if so, completes
** the passed SCANNODE with MPEG Audio-related information. "filename" is used
** for error messages only.
**
*/
BOOL examine_mpegaudio_file(FILE *file, char *filename, SCANNODE *node)
{
	MPEGAUDIOINFO info;
	unsigned char buf[2048];
	unsigned int header_pos, offset;
	BOOL vbr_detected;
	float vbr_avgbitrate = 0, bytes_per_frame, time_per_frame;

	/* Rewind file to its beginning */
	fseek(file, 0, SEEK_SET);

	/* Clear MPEGAUDIOINFO structure */
	memset(&info, 0, sizeof(MPEGAUDIOINFO));

	/* Check for (and parse) MPEG audio frame header */
	if (!find_mpegaudio_header(file, filename, &info))
		return FALSE;

	/* Anything before the first MP3 frame is metadata (ID3v2 tag or whatever) */
	header_pos = ftell(file) - MPEG_AUDIO_FRAME_HEADERLEN;
	info.mpegsize = node->size - header_pos;

	/* Determine how many bytes we have to read */
	bytes_per_frame = (int)calc_mpegaudio_framelength(&info, info.bitrate);

	/* Read frame body  */
	if (!(fread(buf, (int)bytes_per_frame, 1, file)))
	{
		fprintf(stderr, "%s: error while reading from file \"%s\": %s!\n",
			_prgname, filename, strerror(errno));
		return FALSE;
	}

	/* We'll now check for VBR files using common headers and artificial intelligence */
	vbr_detected = FALSE;

	/* Check for XING / LAME VBR header, whose position depends on the channelmode */
	if ( (info.channelmode == CHANNELMODE_STEREO) ||
	     (info.channelmode == CHANNELMODE_JSTEREO) ) offset = XING_STEREO_OFFSET;
	else offset = XING_MONO_OFFSET;
	if (!strncmp(&buf[offset], XING_ID, strlen(XING_ID)))
	{
		/* Ok, it's a VBR file with a Xing header */
		vbr_detected = TRUE;

		/* Communicate it's a VBR file to functions that interpret the structure */
		info.bitrate = 0;

		/* Check whether the Xing header contains the number of MPEG frames in the file */
		if (buf[offset + XING_FLAGS_OFFSET] & XING_FRAMES_FLAG)
			info.frames = INT32_BE(buf, offset + XING_FRAMES_OFFSET);

		/* Check whether there's also a LAME encoder header */
		if (!strncmp(&buf[offset + LAME_OFFSET], LAME_ID, strlen(LAME_ID)))
		{
			unsigned int vbr_quallevel, q_param = 0, V_param = 0;

			/* Parse LAME version */
			sprintf(info.encoder_string, "LAME %c.%c%c%c",
				buf[offset + LAME_VER_OFFSET],
				buf[offset + LAME_VER_OFFSET + 2],
				buf[offset + LAME_VER_OFFSET + 3],
				buf[offset + LAME_VER_OFFSET + 4]);

			/* Determine -q and -V parameters */
			vbr_quallevel = buf[offset + LAME_QUALLEVEL_OFFSET];
			if (vbr_quallevel < 100)
			{
				V_param = (100 - vbr_quallevel) / 10;
				q_param = (100 - vbr_quallevel - (10 * V_param));

				sprintf(&info.encoder_string[9], " (-q %d -V %d)",
					q_param, V_param);
			}
		}
		else strcpy(info.encoder_string, "Xing");
	}
	/* Check for Frauenhofer VBR header */
	else if (!strncmp(&buf[FRAUENHOFER_OFFSET], FRAUENHOFER_ID, strlen(FRAUENHOFER_ID)))
	{
		/* Ok, it's a VBR file with the Frauenhofer header (eg. made by Musicmatch Jukebox) */
		vbr_detected = TRUE;

		/* Communicate it's a VBR file to functions that interpret the structure */
		info.bitrate = 0;

		/* Read number of MPEG frames from the Frauenhofer header */
		info.frames = INT32_BE(buf, FRAUENHOFER_FRAMES_OFFSET);

		/* Parse quality setting */
		sprintf(info.encoder_string, "a Frauenhofer-based encoder (quality level %d)",
			buf[FRAUENHOFER_QUALLEVEL_OFFSET]);
	}
	/* Check for LAME tag in CBR files */
	else if ( (info.bitrate) &&
	          (!strncmp(&buf[offset], LAME_CBR_ID, strlen(LAME_CBR_ID))) &&
	          (!strncmp(&buf[offset + LAME_OFFSET], LAME_ID, strlen(LAME_ID))) )
	{
		unsigned int cbr_quallevel, q_param = 0;

		/* Parse LAME version */
		sprintf(info.encoder_string, "LAME %c.%c%c%c",
			buf[offset + LAME_VER_OFFSET],
			buf[offset + LAME_VER_OFFSET + 2],
			buf[offset + LAME_VER_OFFSET + 3],
			buf[offset + LAME_VER_OFFSET + 4]);

		/* Determine -q parameter */
		cbr_quallevel = buf[offset + LAME_QUALLEVEL_OFFSET];
		if (cbr_quallevel < 100)
		{
			while (cbr_quallevel % 10 != 0)
			{
				q_param++;
				cbr_quallevel++;
			}

			sprintf(&info.encoder_string[9], " (-q %d)",
				q_param);
		}
	}

	/* If we didn't detect a VBR file yet, it's either a CBR file or a VBR file
	   without either Xing or Frauenhofer header. We gonna detect this next. */
	if (!vbr_detected)
	{
		MPEGAUDIOINFO next_info;
		int frames_to_read, i;

		/* Use first frame's bitrate for frame body skipping below */
		next_info.bitrate = info.bitrate;

		/* Take first frame's bitrate into account for average bitrate calculation below */
		vbr_avgbitrate = next_info.bitrate;

		/* We check whether the first VBR_DETECT_FRAMES frames have the same bitrate. We
		   may miss headerless VBRs with a lot of silence in the beginning, of course,
		   but then again we don't want to waste too much time on these seldom MP3s.
		   If, however, we detect it's a headerless VBR file, we read VBR_AVGCALC_FRAMES
		   frames instead to get an accurate average bitrate value. */
		frames_to_read = VBR_DETECT_FRAMES;

		for (i = 0; i < frames_to_read; i++)
		{
			/* Clear MPEGAUDIOINFO structure */
			memset(&next_info, 0, sizeof(MPEGAUDIOINFO));

			/* Search for next frame from current position */
			if (!find_mpegaudio_header(file, filename, &next_info))
				break;

			/* MPEG version, MPEG layer, sampling frequency and channelmode must be the
			   same as in the first frame, otherwise this ain't a valid audio frame header */
			if ( (next_info.mpeg_version != info.mpeg_version) ||
			     (next_info.mpeg_layer   != info.mpeg_layer)   ||
			     (next_info.samplingfreq != info.samplingfreq) ||
			     (next_info.channelmode  != info.channelmode) )
			     continue;

			/* Check whether this header has a bitrate different from the first frame */
			if (next_info.bitrate != info.bitrate)
			{
				vbr_detected = TRUE;
				frames_to_read = VBR_AVGCALC_FRAMES;
			}

			/* Regardless whether it was different, this frame counts for the average bitrate */
			vbr_avgbitrate += next_info.bitrate;

			/* Skip frame body */
			bytes_per_frame = (int)calc_mpegaudio_framelength(&info, next_info.bitrate);
			fseek(file, (int)bytes_per_frame, SEEK_CUR);
		}

		/* So, did we detect a VBR file? */
		if (vbr_detected)
		{
			/* Calculate average bitrate and new bytes_per_frame value, taking into account that
			   we need to include the 4 bytes for the frame header in this calculation */
			vbr_avgbitrate /= (frames_to_read + 1);
			bytes_per_frame = calc_mpegaudio_framelength(&info, vbr_avgbitrate) +
			                  MPEG_AUDIO_FRAME_HEADERLEN;

			/* Communicate it's a VBR file to functions that interpret the structure */
			info.bitrate = 0;
		}
	}

	/* Check whether ID3v1 tag is present */
	fseek(file, -(ID3V1_LEN), SEEK_END);
	if (fread(buf, ID3V1_LEN, 1, file) == 1)
	{
		if (!strncmp(buf, ID3V1_ID, strlen(ID3V1_ID)))
		{
			/* Copy information from ID3 tag */
			if (buf[ID3V1_SONGTITLE_OFFSET])
			{
				const int len = (int)sizeof(info.id3v1_songtitle) - 1;

				strncpy(info.id3v1_songtitle, &buf[ID3V1_SONGTITLE_OFFSET], len);
				info.id3v1_songtitle[len] = '\0';
			}
			if (buf[ID3V1_ARTIST_OFFSET])
			{
				const int len = (int)sizeof(info.id3v1_artist) - 1;

				strncpy(info.id3v1_artist, &buf[ID3V1_ARTIST_OFFSET], len);
				info.id3v1_artist[len] = '\0';
			}
			if (buf[ID3V1_ALBUM_OFFSET])
			{
				const int len = (int)sizeof(info.id3v1_album) - 1;

				strncpy(info.id3v1_album, &buf[ID3V1_ALBUM_OFFSET], len);
				info.id3v1_album[len] = '\0';
			}
			buf[ID3V1_COMMENT_OFFSET] = 0;
			if (buf[ID3V1_YEAR_OFFSET])
				info.id3v1_year = atoi(&buf[ID3V1_YEAR_OFFSET]);
			if ( (buf[ID3V1_GENRE_OFFSET] <= ID3V1_GENRE_MAXID) || (buf[ID3V1_GENRE_OFFSET] == UCHAR_MAX) )
				info.id3v1_genre = buf[ID3V1_GENRE_OFFSET];

			/* ID3v1.1 extension: track number */
			if ( (buf[ID3V1_TRACK_OFFSET - 1] == 0) && (buf[ID3V1_TRACK_OFFSET] > 0) )
				info.id3v1_track = buf[ID3V1_TRACK_OFFSET];

			/* The tag should not be counted as MP3 data */
			info.mpegsize -= ID3V1_LEN;
		}
	}

	/* Unless the file is a VBR file with a Xing or Frauenhofer header, we still need to
	   calculate the number of MPEG frames on our own */
	if (!(vbr_detected && info.frames))
		info.frames = (int) ( (float)info.mpegsize / bytes_per_frame );

	/* In preparation, we calculate time_per_frame as samples_per_frame / sampling_frequency
	   where samples_per_frame is defined in [ISO11172-3], page 20, and mpegaudio.h. */
	if (info.mpeg_layer == MPEG_AUDIO_LAYER1) time_per_frame = L1_TPF_MULT;
	else time_per_frame = L2_L3_TPF_MULT;
	time_per_frame = time_per_frame / (float)info.samplingfreq;

	/* And finally we can calculate song_length as time_per_frame * #frames */
	info.length = time_per_frame * info.frames;

	/* Now all that is left to do is to store a copy of our completed MPEGAUDIOINFO
	   structure along with this file's SCANNODE. */
	node->info = malloc(sizeof(MPEGAUDIOINFO));
	if (!node->info)
	{
		fprintf(stderr, "%s: out of memory while processing file \"%s\": %s!\n",
			_prgname, filename, strerror(errno));
		return FALSE;
	}
	memcpy(node->info, &info, sizeof(MPEGAUDIOINFO));
	node->infotype = INFOTYPE_MPEGAUDIO;

	return TRUE;
}

/*
** infostr = get_mpegaudio_info_string(info, buf, len)
**
** Returns an ASCII representation of the specified MPEGAUDIOINFO structure "info"
** in the specified buffer "buf" (which must be sized appropriately).
*/
void get_mpegaudio_info_string(MPEGAUDIOINFO *info, char *buf)
{
	char *p;

	p = buf;

	switch (info->mpeg_version)
	{
		case MPEG_AUDIO_V1:
			p += sprintf(p, "MPEG-1 ");
			break;
		case MPEG_AUDIO_V2:
			p += sprintf(p, "MPEG-2 ");
			break;
	}

	switch (info->mpeg_layer)
	{
		case MPEG_AUDIO_LAYER1:
			p += sprintf(p, "Layer 1: ");
			break;
		case MPEG_AUDIO_LAYER2:
			p += sprintf(p, "Layer 2: ");
			break;
		case MPEG_AUDIO_LAYER3:
			p += sprintf(p, "Layer 3: ");
	}

	if (info->bitrate) p += sprintf(p, "%3d", info->bitrate);
	else p += sprintf(p, "VBR");

	p += sprintf(p, "/%2d ", info->samplingfreq);

	switch (info->channelmode)
	{
		case CHANNELMODE_STEREO:
			p += sprintf(p, "Stereo ");
			break;

		case CHANNELMODE_JSTEREO:
			p += sprintf(p, "JStereo");
			break;

		case CHANNELMODE_DUALMONO:
			p += sprintf(p, "Mono x2");
			break;

		case CHANNELMODE_MONO:
			p += sprintf(p, "Mono   ");
			break;
		default:
			return;
	}

	p += sprintf(p," [%2dm%02ds]  A:", (info->length / 60), (info->length - ((info->length / 60) * 60)) );

	if (*info->id3v1_artist)
		p += sprintf(p, "%-30.30s", info->id3v1_artist);
	else p += sprintf(p, "%-30.30s", "<no ID3 tag>");

	p += sprintf(p, " T:");
	if (*info->id3v1_songtitle)
		p += sprintf(p, "%-30.30s", info->id3v1_songtitle);
	else p += sprintf(p, "%-30.30s", "<no ID3 tag>");

	p += sprintf(p, " R:");
	if (*info->id3v1_album)
		p += sprintf(p, "%-30.30s", info->id3v1_album);
	else p += sprintf(p, "%-30.30s", "<no ID3 tag>");

	if (info->id3v1_year)
		p += sprintf(p, " Y:%4.4d", info->id3v1_year);
	else p += sprintf(p, " Y:????");

	if (info->id3v1_track)
		p += sprintf(p, " T#:%2.2d", info->id3v1_track);
	else p += sprintf(p, " T#:??");

	if ( (*info->id3v1_artist) || (*info->id3v1_songtitle) )
	{
		if (info->id3v1_genre <= ID3V1_GENRE_MAXID)
			p += sprintf(p, " G:%-30.30s", id3v1_genres[info->id3v1_genre]);
		else p += sprintf(p, " G:%-30.30s", "Unknown");
	}
	else p += sprintf(p, " G:%-30.30s", "<no ID3 tag>");

	if (strlen(info->encoder_string))
	{
		sprintf(p, "  encoded using %s", info->encoder_string);
	}
}
