/*
** $Id: waveaudio.c,v 1.3 2003/01/10 00:59:34 pief Exp $
**
** makelist - MetalServe file list generator
** Copyright (c) 2002-2003 by Pieter Hollants <contact@metalserve.org>
**
** This program is licensed under the GNU General Public License, version 2,
** as published by the Free Software Foundation and available in the file
** COPYING and the Internet location http://www.gnu.org/licenses/gpl.html.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
** or FITNESS FOR A PARTICULAR PURPOSE.
**
** Wave Audio information functions
**
** $Log: waveaudio.c,v $
** Revision 1.3  2003/01/10 00:59:34  pief
** Extended copyright note to the new year
**
** Revision 1.2  2002/10/19 22:35:26  pief
** Renamed makelist.[ch] to main.[ch]
**
** Revision 1.1  2002/09/24 21:46:36  pief
** Added first version of Wave Audio related functions
**
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include <errno.h>

#include "../common/base.h"
#include "../common/list.h"

#include "main.h"
#include "waveaudio.h"

extern char *_prgname;

/*
** rc = examine_waveaudio_file(FILE *file, char *filename, SCANNODE *node)
**
** Checks whether the passed file is a Wave audio file and, if so, completes
** the passed SCANNODE with Wave Audio-related information. "filename" is used
** for error messages only.
**
*/
BOOL examine_waveaudio_file(FILE *file, char *filename, SCANNODE *node)
{
	WAVEAUDIOINFO info;
	unsigned char buf[128];
	int len;

	/* Rewind file to its beginning */
	fseek(file, 0, SEEK_SET);

	/* Read the first 12 bytes */
	if (fread(buf, 12, 1, file) != 1)
	{
		if (ferror(file))
		{
			fprintf(stderr, "%s: error while reading from file \"%s\": %s!\n",
				_prgname, filename, strerror(errno));
		}
		return FALSE;
	}

	/* Check for "RIFF" and "WAVE" signatures */
	if ( (strncmp(&buf[RIFF_OFFSET], RIFF_ID, strlen(RIFF_ID))) ||
	     (strncmp(&buf[WAVE_OFFSET], WAVE_ID, strlen(WAVE_ID))) )
		return FALSE;

	/* Check embedded filesize */
	if (INT32_LE(buf, RIFF_LEN_OFFSET) != (node->size - 8))
		return FALSE;

	/* Read the identifer and the size of the hopefully present "fmt " chunk */
	if (fread(buf, 8, 1, file) != 1)
	{
		if (ferror(file))
		{
			fprintf(stderr, "%s: error while reading from file \"%s\": %s!\n",
				_prgname, filename, strerror(errno));
		}
		return FALSE;
	}

	/* Check for "fmt " signature */
	if (strncmp(buf, FMT_ID, strlen(FMT_ID)))
	{
		printf("\"fmt \" signature missing\n");
		return FALSE;
	}

	/* Determine size and read remaining part of "fmt " chunk */
	len = INT32_LE(buf, RIFF_LEN_OFFSET);
	if (fread(buf, (len-8), 1, file) != 1)
	{
		if (ferror(file))
		{
			fprintf(stderr, "%s: error while reading from file \"%s\": %s!\n",
				_prgname, filename, strerror(errno));
		}
		return FALSE;
	}
//	printf("\"fmt \" chunk length: %d\n", len);

	/* Clear WAVEAUDIOINFO structure */
	memset(&info, 0, sizeof(WAVEAUDIOINFO));

	/* Copy data */
	info.format = INT16_LE(buf, FMT_FORMAT_OFFSET);
	info.channels = INT16_LE(buf, FMT_CHANNELS_OFFSET);
	info.samplingfreq = INT32_LE(buf, FMT_SAMPLINGFREQ_OFFSET);
	info.bitspersample = INT16_LE(buf, FMT_BITSPERSAMPLE_OFFSET);

	/* If this Wave file contains uncompressed data, calculate song length */
	if (info.format == FORMAT_ID_PCM)
	{
		
		
	}
	/* Otherwise read the next 8 bytes (hopefully a new chunk) */
	else
	{
		if (fread(buf, 8, 1, file) != 1)
		{
			if (ferror(file))
			{
				fprintf(stderr, "%s: error while reading from file \"%s\": %s!\n",
					_prgname, filename, strerror(errno));
			}
			return FALSE;
		}

		/* Check for "fact" signature */
/*		if (!(strncmp(buf, FACT_ID, strlen(FACT_ID))))
		{*/
	}

	/* Now all that is left to do is to store a copy of our completed WAVEAUDIOINFO
	   structure along with this file's scannode. */
	node->info = malloc(sizeof(WAVEAUDIOINFO));
	if (!node->info)
	{
		fprintf(stderr, "%s: out of memory while processing file \"%s\": %s!\n",
			_prgname, filename, strerror(errno));
		return FALSE;
	}
	memcpy(node->info, &info, sizeof(WAVEAUDIOINFO));
	node->infotype = INFOTYPE_WAVEAUDIO;

	return TRUE;
}

/*
** infostr = get_waveaudio_info_string(info, buf, len)
**
** Returns an ASCII representation of the specified WAVEAUDIOINFO structure "info"
** in the specified buffer "buf" (which must be sized appropriately).
*/
void get_waveaudio_info_string(WAVEAUDIOINFO *info, char *buf)
{
	char *p;

	p = buf;

	p += sprintf(p, "Wave Audio: %d-bit/", info->bitspersample);

	switch (info->channels)
	{
		case 1:
			p += sprintf(p, "Mono");
			break;
		case 2:
			p += sprintf(p, "Stereo");
			break;
		default:
			p += sprintf(p, "%d channels", info->channels);
	}

	p += sprintf(p, "/%dHz ", info->samplingfreq);

	switch (info->format)
	{
		case FORMAT_ID_PCM:
			p += sprintf(p, FORMAT_STR_PCM);
			break;
		case FORMAT_ID_MS_ADPCM:
			p += sprintf(p, FORMAT_STR_MS_ADPCM);
			break;
		case FORMAT_ID_ITU_G711_ALAW:
			p += sprintf(p, FORMAT_STR_ITU_G711_ALAW);
			break;
		case FORMAT_ID_ITU_G711_MULAW:
			p += sprintf(p, FORMAT_STR_ITU_G711_MULAW);
			break;
		case FORMAT_ID_DVI_IMA_ADPCM:
			p += sprintf(p, FORMAT_STR_DVI_IMA_ADPCM);
			break;
		case FORMAT_ID_DOLBY_AC2:
			p += sprintf(p, FORMAT_STR_DOLBY_AC2);
			break;
		case FORMAT_ID_MPEG:
			p += sprintf(p, FORMAT_STR_MPEG);
			break;
		case FORMAT_ID_MP3:
			p += sprintf(p, FORMAT_STR_MP3);
			break;
		case FORMAT_ID_DOLBY_AC3_SPDIF:
			p += sprintf(p, FORMAT_STR_DOLBY_AC3_SPDIF);
			break;
		default:
			p += sprintf(p, "Unsupported format %d", info->format);
	}
}
