/*
** $Id: main.c,v 1.1 2003/01/10 01:02:38 pief Exp $
**
** metalserved - IRC DCC file server
** Copyright (c) 2002-2003 by Pieter Hollants <contact@metalserve.org>
**
** This program is licensed under the GNU General Public License, version 2,
** as published by the Free Software Foundation and available in the file
** COPYING and the Internet location http://www.gnu.org/licenses/gpl.html.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
** or FITNESS FOR A PARTICULAR PURPOSE.
**
** Main program
**
** $Log: main.c,v $
** Revision 1.1  2003/01/10 01:02:38  pief
** Added first version of metalserved, which yet does nothing but full
** configuration file processing, including subsections.
**
**
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include <errno.h>
#include <libgen.h>

#include "../lib/getopt.h"

#include "../common/base.h"
#include "../common/list.h"
#include "../common/conf.h"
#include "../common/log.h"
#include "../common/misc.h"
#include "../common/strings.h"

#include "main.h"
#include "servers.h"

const char *_prgbanner =
	"%s [%s v%s] - %s IRC DCC file server\n"
	"Copyright (c) 2002 by Pieter Hollants <contact@metalserve.org> - http://www.metalserve.org\n\n";

char *_prgname;

BOOL _call_print_config = FALSE;

/*
** Prototypes for functions referenced before they're defined (forward declarations)
*/
APFUNC(argparser_network);
APFUNC(argparser_server);
APFUNC(argparser_channel);
static void print_config(void);

/*
** Command line options
*/
static struct option long_options[] =
{
	{ "configfile",		required_argument,	0, 'f' },
	{ "help",		no_argument,		0, 'h' },
	{ "serverroot",		required_argument,	0, 's' },
	{ "test-config",	no_argument,		0, 't' },
	{ "verbose",		no_argument,		0, 'v' },
	{ "version",		no_argument,		0, 'V' },
	{ 0, 0, 0, 0 }
};
const char *_cmdlineopts = "fhqstvV";

const char *_cmdlinehelp =
	"This program is licensed under the GNU General Public License, version 2.\n"
	"See the file COPYING or visit http://www.gnu.org/licenses/gpl.html for details.\n\n"

	"Usage: %s [options]\n\n"

	"Options:\n"
	"  -f, --configfile=FILE           specify alternative configuration file\n"
	"  -h, --help                      display this information\n"
	"  -s, --serverroot=PATH           override configured serverroot (root directory for file serving)\n"
	"  -t, --test-config               test configuration file\n"
	"  -v, --verbose                   increase verboseness (loglevel)\n"
	"  -V, --version                   print version and exit\n";

/*
** Global configuration variables
*/
char *_cfgfile_name;

char *_logfile_name;
static int _loglevel		= DEF_LOGLEVEL;
char *_serverroot;
char **_nickname;
char *_realname;
char *_email_address;

int _server_connect_timeout	= DEF_SERVER_CONNECT_TIMEOUT;
int _server_timeout		= DEF_SERVER_TIMEOUT;
int _part_timeout		= DEF_PART_TIMEOUT;
BOOL _part_finishdccs		= DEF_PART_FINISHDCCS;
BOOL _client_lookups		= DEF_CLIENT_LOOKUPS;
char **_ignored_users;
int _message_linedelay		= DEF_MSG_LINEDELAY;
int _max_find_results		= DEF_MAX_FIND_RESULTS;
int _dcc_minport		= DEF_DCC_MINPORT,
    _dcc_maxport		= DEF_DCC_MAXPORT;
int _dcc_timeout		= DEF_DCC_TIMEOUT;
int _dcc_maxxfers		= DEF_DCC_MAXXFERS,
    _dcc_maxqueue		= DEF_DCC_MAXQUEUE;
int _dcc_maxuserxfers		= DEF_DCC_MAXUSERXFERS,
    _dcc_maxuserqueue		= DEF_DCC_MAXUSERQUEUE;
int _dcc_maxlistxfers		= DEF_DCC_MAXLISTXFERS,
    _dcc_maxlistqueue		= DEF_DCC_MAXLISTQUEUE;
int _dcc_mincps			= DEF_DCC_MINCPS;
int _dcc_largefiles_threshold	= DEF_DCC_LARGEFILES_THRESHOLD,
    _dcc_largefiles_mincps	= DEF_DCC_LARGEFILES_MINCPS;

/* Valid arguments and associated values to the "loglevel" configuration directive. We intentionally
   don't try to get the strings from syslog.h, since this is what the user specifies in the config
   file and that should be under our control. */
CFGSTRVALUE _loglevelargs[] =
{
        /* string,      value,          exclusive? */
        { "critical",   LOG_CRIT,       TRUE },
        { "error",      LOG_ERR,        TRUE },
        { "warning",    LOG_WARNING,    TRUE },
        { "notice",     LOG_NOTICE,     TRUE },
        { "info",       LOG_INFO,       TRUE },
        { "debug",      LOG_DEBUG,      TRUE },
        LAST_CFGSTRVALUE
};

/* Configuration directives (see src/common/conf.[ch]) */
CFGDIRECTIVE directives[] =
{
	/* SUBSECTION_DIRECTIVE(name, 				argparser function,
				valid_sections,			must_open, target_section) */
	SUBSECTION_DIRECTIVE("network",				argparser_network, \
			     GLOBAL_SECTION,			TRUE,	NETWORK_SECTION),
	SUBSECTION_DIRECTIVE("server",				argparser_server, \
			     GLOBAL_SECTION | NETWORK_SECTION,	FALSE,	SERVER_SECTION),
	SUBSECTION_DIRECTIVE("channel",				argparser_channel, \
			     GLOBAL_SECTION | NETWORK_SECTION,	FALSE,	CHANNEL_SECTION),

	/* directive name,		argparser function,			target variable,
	   min, max,			section,				params */
	{ "logfile",			argparser_string,			&_logfile_name,
	  0, FILENAME_MAX,		GLOBAL_SECTION,				NULL },
	{ "loglevel",			argparser_stringlist,			&_loglevel,
	  0, 1,				GLOBAL_SECTION,				_loglevelargs },
	{ "serverroot",			argparser_string,			&_serverroot,
	  0, PATH_MAX,			GLOBAL_SECTION,				NULL },
	{ "nickname",			argparser_stringlist,			&_nickname,
	  0, MAX_VALIST_PTRS,		GLOBAL_SECTION,				NULL },
	{ "realname",			argparser_string,			&_realname,
	  0, MAX_REALNAME_LEN,		GLOBAL_SECTION,				NULL },
	{ "email_address",		argparser_string,			&_email_address,
	  0, MAX_EMAILADDR_LEN,		GLOBAL_SECTION,				NULL },
	{ "advertise_delay",		argparser_int,				offsetof(CHANNEL, advertise_delay),
	  0, MAX_ADVERTISE_DELAY,	CHANNEL_SECTION,			NULL },
	{ "find_trigger",		argparser_switch,			offsetof(CHANNEL, find_trigger),
	  0, 0,				CHANNEL_SECTION,			NULL },
	{ "priority",			argparser_int,				offsetof(PRIONODE, priority),
	  MINPRI, MAXPRI,		CHANNEL_SECTION | SERVER_SECTION,	NULL },
	{ "ports",			argparser_intrange,			offsetof(SERVER, minport),
	  1, 65535,			SERVER_SECTION,				NULL },
	{ "password",			argparser_string,			offsetof(SERVER, password),
	  0, MAX_PASSWORD_LEN,		SERVER_SECTION,				NULL },
	{ "server_connect_timeout",	argparser_int,				&_server_connect_timeout,
	  0, MAX_TIMEOUT,		GLOBAL_SECTION,				NULL },
	{ "server_timeout",		argparser_int,				&_server_timeout,
	  0, MAX_TIMEOUT,		GLOBAL_SECTION,				NULL },
	{ "part_timeout",		argparser_int,				&_part_timeout,
	  0, MAX_TIMEOUT,		GLOBAL_SECTION, 			NULL },
	{ "part_finishdccs",		argparser_switch,			&_part_finishdccs,
	  0, 0,				GLOBAL_SECTION,				NULL },
	{ "client_lookups",		argparser_switch,			&_client_lookups,
	  0, 0,				GLOBAL_SECTION,				NULL },
	{ "message_linedelay",		argparser_int,				&_message_linedelay,
	  0, MAX_MSG_LINEDELAY,		GLOBAL_SECTION,				NULL },
	{ "max_find_results",		argparser_int,				&_max_find_results,
	  0, MAX_MAX_FIND_RESULTS,	GLOBAL_SECTION, 			NULL },
	{ "dcc_ports",			argparser_intrange,			&_dcc_minport,
	  MIN_DCC_PORT, MAX_DCC_PORT,	GLOBAL_SECTION,				NULL },
	{ "dcc_timeout",		argparser_int,				&_dcc_timeout,
	  0, MAX_TIMEOUT,		GLOBAL_SECTION,				NULL },
	{ "dcc_maxxfers",		argparser_int,				&_dcc_maxxfers,
	  0, MAX_XFERS,			GLOBAL_SECTION,				NULL },
	{ "dcc_maxqueue",		argparser_int,				&_dcc_maxqueue,
	  0, MAX_QUEUE,			GLOBAL_SECTION, 			NULL },
	{ "dcc_maxuserxfers",		argparser_int,				&_dcc_maxuserxfers,
	  0, MAX_XFERS,			GLOBAL_SECTION, 			NULL },
	{ "dcc_maxuserqueue",		argparser_int,				&_dcc_maxuserqueue,
	  0, MAX_QUEUE,			GLOBAL_SECTION,				NULL },
	{ "dcc_maxlistxfers",		argparser_int,				&_dcc_maxlistxfers,
	  0, MAX_XFERS,			GLOBAL_SECTION,				NULL },
	{ "dcc_maxlistqueue",		argparser_int,				&_dcc_maxlistqueue,
	  0, MAX_QUEUE,			GLOBAL_SECTION, 			NULL },
	{ "dcc_mincps",			argparser_int,				&_dcc_mincps,
	  0, MAX_CPS,			GLOBAL_SECTION, 			NULL },
	{ "dcc_largefiles_threshold",	argparser_int,				&_dcc_largefiles_threshold,
	  0, MAX_LARGEFILES_THRESHOLD,	GLOBAL_SECTION,				NULL },
	{ "dcc_largefiles_mincps",	argparser_int,				&_dcc_largefiles_mincps,
	  0, MAX_CPS,			GLOBAL_SECTION,				NULL },
	LAST_CFGDIRECTIVE
};

/* Internal data structures */
SERVER _defserver;
CHANNEL _defchannel;

LIST *_networks;

/*
** rc = parse_cmdline_args(argc, argv)
**
** Parses the program's command line arguments.
**
** Returns TRUE if the program should continue to run or FALSE if the
** program should quit.
*/
static BOOL parse_cmdline_args(int argc, char **argv)
{
	int c, option_index = 0;

	while (TRUE)
	{
		c = getopt_long(argc, argv, _cmdlineopts, long_options, &option_index);
		if (c == -1) break;

		switch (c)
		{
			case 'f':
			{
				_cfgfile_name = strdup(optarg);
				break;
			}
			case 'h':
			{
				printf(_prgbanner, _prgname, PACKAGE_NAME, PACKAGE_VERSION, PACKAGE_NAME);
				printf(_cmdlinehelp, _prgname);
				return FALSE;
			}
			case 's':
			{
				_serverroot = strdup(optarg);
				break;
			}
			case 't':
			{
				_call_print_config = TRUE;
				break;
			}
			case 'v':
			{
				_loglevel++;
				break;
			}
			case 'V':
			{
				printf(_prgbanner, _prgname, PACKAGE_NAME, PACKAGE_VERSION, PACKAGE_NAME);
				printf("Compiled %s %s\n", __DATE__, __TIME__);
				return FALSE;
			}
			case '?':
			{
				/* getopt_long already printed an error message */
				return FALSE;
			}
		}
	}

	return TRUE;
}

/*
** rc = parse_cfgfile()
**
** Parses the MetalServe configuration file.
**
** Returns TRUE if the program should continue to run or FALSE if the
** program should quit.
*/
static BOOL parse_cfgfile(void)
{
	int attempt;
	CFGFILE *cfgfile = NULL;
	char *errstr;
	char *old_serverroot;
	int old_loglevel;

	/* Attempt to open configuration file at one of two locations */
	for (attempt = 0; attempt < 2; attempt++)
	{
		/* Try the system-wide location first */
		if (attempt == 0)
			_cfgfile_name = DEF_CFGFILE_NAME;

		/* Then try the home dir of the user we're running under */
		else
		{
			char *homedir;

			/* Get path of current user's home directory and allocate memory for composed filename */
			homedir = getenv("HOME");
			_cfgfile_name = malloc(PATH_MAX);
			if ( (!homedir) || (!_cfgfile_name) )
			{
				fprintf(stderr, "%s: could not compose full filename of \"%s\" configuration file!\n",
					_prgname, HOMEDIR_CFGFILE_NAME);
				return FALSE;
			}
			
			snprintf(_cfgfile_name, PATH_MAX, "%s/%s", homedir, HOMEDIR_CFGFILE_NAME);
		}

		/* Now try to actually open the configuration file */
		errno = 0;
		cfgfile = cfgopen(_cfgfile_name, CFGOPENFLAGS_IGNUNKNOWN);
		if ( (cfgfile) || ( (!cfgfile) && (errno != ENOENT) ) )
			break;
	}
	
	/* Couldn't we open it for some reason? */
	if (!cfgfile)
	{
		/* File did not exist at both locations */
		if (errno == ENOENT)
		{
			fprintf(stderr, "%s: missing configuration file at either \"%s\" or \"%s\"!\n",
				_prgname, DEF_CFGFILE_NAME, HOMEDIR_CFGFILE_NAME);
		}
		/* Some other serious error */
		else
		{	
			fprintf(stderr, "%s: error opening configuration file \"%s\":\n%s!\n",
				_prgname, _cfgfile_name, strerror(errno));
		}

		return FALSE;
	}

	/* Process configuration directives (but preserve variables eventually modified on command line) */
	old_serverroot = _serverroot;
	old_loglevel = _loglevel;
	_loglevel = 0;
	if (!cfgread(cfgfile, directives))
	{
		if (cfgerrno(cfgfile)) errstr = cfgerrstr(cfgfile);
		else errstr = strerror(errno);
		fprintf(stderr, "%s: error while reading line %d of \"%s\":\n%s!\n",
			_prgname, cfglineno(cfgfile), _cfgfile_name, errstr);
		return FALSE;
	}
	if (old_serverroot)
		_serverroot = old_serverroot;
	if (_loglevel)
	{
		_loglevel += (old_loglevel - DEF_LOGLEVEL);
		if (_loglevel > LOG_DEBUG)
			_loglevel = LOG_DEBUG;
	}
	else
		_loglevel = old_loglevel;

	cfgclose(cfgfile);

	return TRUE;
}

/*
** Argument parser function for the "network" directive
*/
APFUNC(argparser_network)
{                               
	NETWORK *network;

	/* As long as there are no configuration directives in the "network" section,
	   specifiying a network with the special keyword "default" will yield an
	   error. */
	if (!strcasecmp(*argp, "default"))
		return CFG_ERR_INVDEFARG;

	/* Check whether the specified network already exists */
	for (network = (NETWORK *)_networks->first; network; network = (NETWORK *)network->n.next)
	{
		if (!strcasecmp(network->name, *argp))
			return CFG_ERR_DUPLICATEARG;
	}

	/* Allocate and initialize new NETWORK structure */
	network = (NETWORK *) allocate_network();
	if (!network)
		return CFG_ERR_OUTOFMEM;

	network->name = unescquote(*argp);
	if (!network->name)
	{
		free_network(network);
		return CFG_ERR_OUTOFMEM;
	}

	/* Append structure to list of NETWORK structure */
	append_node(_networks, (NODE *)network);

	/* Set structure base pointer */
	*((void **)target) = network;

	return CFG_ERR_NONE;
}
                
/*
** Argument parser function for the "server" directive
*/
APFUNC(argparser_server)
{
	NETWORK *network;
	SERVER *server;

	/* Check whether this "server" directive specifies default properties
	   for all servers */
	if (!strcasecmp(*argp, "default"))
	{
		/* Default properties can be specified in the global section only */
		if (section != GLOBAL_SECTION)
			return CFG_ERR_INVSECTION;

		/* Set structure base pointer to structure for default server */
		*((void **)target) = &_defserver;

		return CFG_ERR_NONE;
	}

	/* Check whether this directive specifies properties for a server without
	   being located inside a NETWORK section */
	network = *((NETWORK **)target);
	if (!network)
		return CFG_ERR_INVSECTION;

	/* Check whether the specified server already exists */
	for (server = (SERVER *)network->servers->first; server; server = (SERVER *)server->n.next)
	{
		if (!strcasecmp(server->conn.name, *argp))
			return CFG_ERR_DUPLICATEARG;
	}

	/* Check length of server name */
	if (strlen(*argp) > MAX_SERVER_LEN)
		return CFG_ERR_ARGTOOLONG;

	/* Allocate and initialize new SERVER structure */
	server = (SERVER *)allocate_server();
	if (!server)
		return CFG_ERR_OUTOFMEM;

	server->conn.name = unescquote(*argp);
	if (!server->conn.name)
	{
		free_server(server);
		return CFG_ERR_OUTOFMEM;
	}

	/* Append structure to network's list of SERVER structures */
	append_node(network->servers, (NODE *)server);

	/* Set structure base pointer */
	*((void **)target) = server;

	return CFG_ERR_NONE;
}
                 
/*
** Argument parser function for the "channel" directive
*/
APFUNC(argparser_channel)
{
	NETWORK *network;
	CHANNEL *channel;

	/* Check whether this "channel" directive specifies default properties
	   for all channels */
	if (!strcasecmp(*argp, "default"))
	{
		/* Default properties can be specified in the global section only */
		if (section != GLOBAL_SECTION)
			return CFG_ERR_INVSECTION;

		/* Set structure base pointer to structure for default channel */
		*((void **)target) = &_defchannel;

		return CFG_ERR_NONE;
	}

	/* Check whether this directive specifies properties for a channel without
	   being located inside a NETWORK section */
	network = *((NETWORK **)target);
	if (!network)
		return CFG_ERR_INVSECTION;

	/* Check whether the specified channel already exists */
	for (channel = (CHANNEL *)network->channels->first; channel; channel = (CHANNEL *)channel->n.next)
	{
		if (!strcasecmp(channel->name, *argp))
			return CFG_ERR_DUPLICATEARG;
	}

	/* Check length of channel name */
	if (strlen(*argp) > MAX_CHANNEL_LEN)
		return CFG_ERR_ARGTOOLONG;

	/* Allocate and initialize new CHANNEL structure */
	channel = (CHANNEL *)allocate_channel();
	if (!channel)
		return CFG_ERR_OUTOFMEM;

	channel->name = unescquote(*argp);
	if (!channel->name)
	{
		free_channel(channel);
		return CFG_ERR_OUTOFMEM;
	}

	/* Append structure to network's list of CHANNEL structures */
	append_node(network->channels, (NODE *)channel);

	/* Set structure base pointer */
	*((void **)target) = channel;

	return CFG_ERR_NONE;
}

/*
** print_config()
**
** Dumps the current configuration to stdout. Invoked when metalserved is
** started with the "-t" (test configfile) argument.
*/
static void print_config(void)
{
	char **strlistentry;
	char buf[12];
	NETWORK *network;
	SERVER *server;
	CHANNEL *channel;

	printf("%s configuration dump:\n\n", _prgname);

	printf("I. Global settings\n\n");

	printf("Directive                Value                     Description\n");
	printf("---------------------------------------------------------------------------------------------------\n");

	printf("logfile                  %-24.24s  Destination for log messages\n",
	       (_logfile_name ? _logfile_name : "(disabled)"));
	printf("loglevel                 %-24.24s  Controls the amount of messages logged\n\n",
	       getloglevelname(_loglevel));

	printf("serverroot               %-24.24s  Root directory from which files are served\n\n",
	       _serverroot);

	for (strlistentry = _nickname; *strlistentry; strlistentry++)
	{
		if (strlistentry == _nickname)
			printf("nickname                 %-24.24s  The Nickname used on IRC\n",
			       *strlistentry);
		else
			printf("                         %s\n", *strlistentry);
	}

	printf("realname                 %-24.24s  Shown when somebody /WHOIS's us\n",
	       _realname);
	printf("email_address            %-24.24s  Shown when somebody /WHOIS's us\n\n",
	       _email_address);

	printf("server_connect_timeout   %-24d  Timeout for establishing connections to servers\n",
	       _server_connect_timeout);
	printf("server_timeout           %-24d  Timeout for established server connections\n",
	       _server_timeout);
	printf("part_timeout             %-24d  Time a user may be missing from a channel\n",
	       _part_timeout);
	printf("part_finishdccs          %-24s  Finish transfers to PARTed (missing) users?\n",
	       (_part_finishdccs ? "Enabled" : "Disabled"));
	printf("client_lookups           %-24s  Look up IP addresses of connecting clients?\n",
	       (_client_lookups ? "Enabled" : "Disabled"));

	if (_ignored_users)
	{
		for (strlistentry = _ignored_users; *strlistentry; strlistentry++)
		{
			if (strlistentry == _ignored_users)
				printf("ignored_users            %-24.24s  Users whose requests will be ignored\n",
				       *strlistentry);
			else
				printf("                         %s\n", *strlistentry);
		}
	}
	else
		printf("ignored_users            %-24.24s  Users whose requests will be ignored\n", "(None)");
	
	printf("message_linedelay        %-24d  Delay in ms between each output line\n",
	       _message_linedelay);
	printf("max_find_results         %-24d  Maximum #results to @find/@locator requests\n\n",
	       _max_find_results);

	sprintf(buf, "%d-%d", _dcc_minport, _dcc_maxport);
	printf("dcc_ports                %-24s  TCP port range for DCC SENDs\n", buf);
	printf("dcc_timeout              %-24d  Timeout for DCC SENDs\n",
	       _dcc_timeout);
	printf("dcc_maxxfers             %-24d  Max. number of concurrent xfers\n",
	       _dcc_maxxfers);
	printf("dcc_maxqueue             %-24d  Number of queue slots\n",
	       _dcc_maxqueue);
	printf("dcc_maxuserxfers         %-24d  Max. number of concurrent xfers per user\n",
	       _dcc_maxuserxfers);
	printf("dcc_maxuserqueue         %-24d  Max. number of queue slots per user\n",
	       _dcc_maxuserqueue);
	printf("dcc_maxlistqueue         %-24d  Max. number of concurrent xfers of filelist\n",
	       _dcc_maxlistqueue);
	printf("dcc_maxlistqueue         %-24d  Max. number of queue slots for filelist\n\n",
	       _dcc_maxlistqueue);
 
	printf("dcc_mincps               %-24d  Minimum required CPS for ordinary xfers\n",
	       _dcc_mincps);
	printf("dcc_largefiles_threshold %-24d  Threshold in KB for a file to be a 'large file'\n",
	       _dcc_largefiles_threshold);
	printf("dcc_largefiles_mincps    %-24d  Minimum required CPS for xfers of 'large files'\n\n",
	       _dcc_largefiles_mincps);

	printf("II. IRC networks\n\n");

	for (network = (NETWORK *)_networks->first; network; network = (NETWORK *)network->n.next)
	{
		printf("Network %s\n", network->name);

		for (server = (SERVER *)network->servers->first; server; server = (SERVER *)server->n.next)
		{
			sprintf(buf, "%d-%d", server->minport, server->maxport);
			printf("  Server %s (priority %d, ports %s, ",
			       server->conn.name, server->n.priority, buf);
			if (server->password)
				printf("password \"%s\")\n", server->password);
			else
				printf("no password configured)\n");
		}

		for (channel = (CHANNEL *)network->channels->first; channel; channel = (CHANNEL *)channel->n.next)
		{
			printf("  Channel %s (priority %d, advertise delay %d secs, find_trigger %s)\n",
			       channel->name, channel->n.priority, channel->advertise_delay,
			       (channel->find_trigger ? "enabled" : "disabled"));
		}

		printf("\n");
	}

	printf("Note: this dump includes only directives relevant to %s. Call makelist\n"
	       "with the -t argument to see the values of additional directives.\n", _prgname);
}

/*
** Main routine
*/
int main(int argc, char **argv)
{
	BOOL rc;

	/* Determine program name */
	_prgname = (char *)basename(strdup(argv[0]));

	/* Set up default server and channel structures */
	_defserver.minport          = DEF_SERVER_MINPORT;
	_defserver.maxport          = DEF_SERVER_MAXPORT;
	_defchannel.advertise_delay = DEF_CHANNEL_ADVERTISEDELAY;
	_defchannel.find_trigger    = DEF_CHANNEL_FINDTRIGGER;

	/* Set up list for NETWORK nodes */
	_networks = allocate_list();
	if (!_networks)
	{
		fprintf(stderr, "%s: out of memory while allocating list for NETWORK nodes!",
		        _prgname);
		return 1;
	}

	/* Parse command line arguments */
	if (!parse_cmdline_args(argc, argv))
		return 1;

	/* Parse configuration file */
	if (!parse_cfgfile())
		return 1;

	/* Check whether the -t option was given */
	if (_call_print_config)
	{
		printf(_prgbanner, _prgname, PACKAGE_NAME, PACKAGE_VERSION, PACKAGE_NAME);
		print_config();
		return 0;
	}

        /* Check whether a logfile has been specified */
        if (_logfile_name)
        {
                /* If so, open it */
                if (logopen(_logfile_name, _loglevel))
                {
                        logmsg(LOG_NOTICE, "%s [%s v%s] starting\n", _prgname,
                               PACKAGE_NAME, PACKAGE_VERSION);
                }
                else
                {
                        fprintf(stderr, "%s: could not open logfile \"%s\" for writing:\n%s!\n",
                                _prgname, _logfile_name, strerror(errno));
                        return FALSE;
                }
        }       

	/* Log exit message */
	logmsg(LOG_NOTICE, "terminating\n");

	return 0;
}
